import logging
import time
from typing import List, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.gate_io_spot_candles import constants as CONSTANTS
from hummingbot.logger import HummingbotLogger


class GateioSpotCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)

    @property
    def name(self):
        return f"gate_io_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        return trading_pair.replace("-", "_")

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return False

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return False

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST) -> dict:
        """
        For API documentation, please refer to:
        https://www.gate.io/docs/developers/apiv4/en/#market-candlesticks

        This API only accepts a limit of 10000 candles ago.
        """
        candles_ago = (int(time.time()) - start_time) // self.interval_in_seconds
        if candles_ago > CONSTANTS.MAX_CANDLES_AGO:
            raise ValueError("Gate.io REST API does not support fetching more than 10000 candles ago.")
        return {
            "currency_pair": self._ex_trading_pair,
            "interval": self.interval,
            "from": start_time,
            "to": end_time
        }

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        new_hb_candles = []
        for i in data:
            timestamp = self.ensure_timestamp_in_seconds(i[0])
            open = i[5]
            high = i[3]
            low = i[4]
            close = i[2]
            volume = i[6]
            quote_asset_volume = i[1]
            # no data field
            n_trades = 0
            taker_buy_base_volume = 0
            taker_buy_quote_volume = 0
            new_hb_candles.append([timestamp, open, high, low, close, volume,
                                   quote_asset_volume, n_trades, taker_buy_base_volume,
                                   taker_buy_quote_volume])
        return new_hb_candles

    def ws_subscription_payload(self):
        return {
            "time": int(self._time()),
            "channel": CONSTANTS.WS_CANDLES_ENDPOINT,
            "event": "subscribe",
            "payload": [self.interval, self._ex_trading_pair]
        }

    def _parse_websocket_message(self, data: dict):
        candles_row_dict = {}
        if data.get("event") == "update" and data.get("channel") == "spot.candlesticks":
            candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(data["result"]["t"])
            candles_row_dict["open"] = data["result"]["o"]
            candles_row_dict["high"] = data["result"]["h"]
            candles_row_dict["low"] = data["result"]["l"]
            candles_row_dict["close"] = data["result"]["c"]
            candles_row_dict["volume"] = data["result"]["a"]
            candles_row_dict["quote_asset_volume"] = data["result"]["v"]
            candles_row_dict["n_trades"] = 0
            candles_row_dict["taker_buy_base_volume"] = 0
            candles_row_dict["taker_buy_quote_volume"] = 0
            return candles_row_dict



import argparse
from typing import TYPE_CHECKING, List

from hummingbot.client.command.connect_command import OPTIONS as CONNECT_OPTIONS
from hummingbot.exceptions import ArgumentParserError

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication


class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)

    def exit(self, status=0, message=None):
        pass

    def print_help(self, file=None):
        pass

    @property
    def subparser_action(self):
        for action in self._actions:
            if isinstance(action, argparse._SubParsersAction):
                return action

    @property
    def commands(self) -> List[str]:
        return list(self.subparser_action._name_parser_map.keys())

    def subcommands_from(self, top_level_command: str) -> List[str]:
        parser: argparse.ArgumentParser = self.subparser_action._name_parser_map.get(top_level_command)
        if parser is None:
            return []
        subcommands = parser._optionals._option_string_actions.keys()
        filtered = list(filter(lambda sub: sub.startswith("--"), subcommands))
        return filtered


def load_parser(hummingbot: "HummingbotApplication", command_tabs) -> ThrowingArgumentParser:
    parser = ThrowingArgumentParser(prog="", add_help=False)
    subparsers = parser.add_subparsers()

    connect_parser = subparsers.add_parser("connect", help="List available exchanges and add API keys to them")
    connect_parser.add_argument("option", nargs="?", choices=CONNECT_OPTIONS, help="Name of the exchange that you want to connect")
    connect_parser.set_defaults(func=hummingbot.connect)

    create_parser = subparsers.add_parser("create", help="Create a new bot")
    create_parser.add_argument("--v2-config", dest="script_to_config", nargs="?", default=None, help="Name of the v2 strategy (from conf/scripts/)")
    create_parser.add_argument("--controller-config", dest="controller_name", nargs="?", default=None, help="Name of the controller")
    create_parser.set_defaults(func=hummingbot.create)

    import_parser = subparsers.add_parser("import", help="Import an existing bot by loading the configuration file")
    import_parser.add_argument("file_name", nargs="?", default=None, help="Name of the configuration file")
    import_parser.set_defaults(func=hummingbot.import_command)

    help_parser = subparsers.add_parser("help", help="List available commands")
    help_parser.add_argument("command", nargs="?", default="all", help="Enter ")
    help_parser.set_defaults(func=hummingbot.help)

    balance_parser = subparsers.add_parser("balance", help="Display your asset balances across all connected exchanges")
    balance_parser.add_argument("option", nargs="?", choices=["limit", "paper"], default=None,
                                help="Option for balance configuration")
    balance_parser.add_argument("args", nargs="*")
    balance_parser.set_defaults(func=hummingbot.balance)

    config_parser = subparsers.add_parser("config", help="Display the current bot's configuration")
    config_parser.add_argument("key", nargs="?", default=None, help="Name of the parameter you want to change")
    config_parser.add_argument("value", nargs="?", default=None, help="New value for the parameter")
    config_parser.set_defaults(func=hummingbot.config)

    start_parser = subparsers.add_parser("start", help="Start the current bot")
    start_parser.add_argument("--v2", type=str, dest="v2_conf",
                              help="V2 strategy config file name (from conf/scripts/)")

    start_parser.set_defaults(func=hummingbot.start)

    stop_parser = subparsers.add_parser('stop', help="Stop the current bot")
    stop_parser.set_defaults(func=hummingbot.stop)

    status_parser = subparsers.add_parser("status", help="Get the market status of the current bot")
    status_parser.add_argument("--live", default=False, action="store_true", dest="live", help="Show status updates")
    status_parser.set_defaults(func=hummingbot.status)

    history_parser = subparsers.add_parser("history", help="See the past performance of the current bot")
    history_parser.add_argument("-d", "--days", type=float, default=0, dest="days",
                                help="How many days in the past (can be decimal value)")
    history_parser.add_argument("-v", "--verbose", action="store_true", default=False,
                                dest="verbose", help="List all trades")
    history_parser.add_argument("-p", "--precision", default=None, type=int,
                                dest="precision", help="Level of precions for values displayed")
    history_parser.set_defaults(func=hummingbot.history)

    lphistory_parser = subparsers.add_parser("lphistory", help="See LP position history and performance")
    lphistory_parser.add_argument("-d", "--days", type=float, default=0, dest="days",
                                  help="How many days in the past (can be decimal value)")
    lphistory_parser.add_argument("-v", "--verbose", action="store_true", default=False,
                                  dest="verbose", help="List all LP position updates")
    lphistory_parser.add_argument("-p", "--precision", default=None, type=int,
                                  dest="precision", help="Level of precision for values displayed")
    lphistory_parser.set_defaults(func=hummingbot.lphistory)

    gateway_parser = subparsers.add_parser("gateway", help="Helper commands for Gateway server.")
    gateway_parser.set_defaults(func=hummingbot.gateway)
    gateway_subparsers = gateway_parser.add_subparsers()

    gateway_allowance_parser = gateway_subparsers.add_parser("allowance", help="Check token allowances for ethereum connectors")
    gateway_allowance_parser.add_argument("connector", nargs="?", default=None, help="Ethereum connector name/type (e.g., uniswap/amm)")
    gateway_allowance_parser.set_defaults(func=hummingbot.gateway_allowance)

    gateway_approve_parser = gateway_subparsers.add_parser("approve", help="Approve token for use with ethereum connectors")
    gateway_approve_parser.add_argument("connector", nargs="?", default=None, help="Connector name/type (e.g., jupiter/router)")
    gateway_approve_parser.add_argument("token", nargs="?", default=None, help="Token symbol to approve (e.g., WETH)")
    gateway_approve_parser.set_defaults(func=hummingbot.gateway_approve)

    gateway_balance_parser = gateway_subparsers.add_parser("balance", help="Check token balances")
    gateway_balance_parser.add_argument("chain", nargs="?", default=None, help="Chain name (e.g., ethereum, solana)")
    gateway_balance_parser.add_argument("tokens", nargs="?", default=None, help="Comma-separated list of tokens to check (optional)")
    gateway_balance_parser.set_defaults(func=hummingbot.gateway_balance)

    gateway_config_parser = gateway_subparsers.add_parser("config", help="Show or update configuration")
    gateway_config_parser.add_argument("namespace", nargs="?", default=None, help="Namespace (e.g., ethereum-mainnet, uniswap)")
    gateway_config_parser.add_argument("action", nargs="?", default=None, help="Action to perform (update)")
    gateway_config_parser.add_argument("args", nargs="*", help="Additional arguments: <path> <value> for direct update")
    gateway_config_parser.set_defaults(func=hummingbot.gateway_config)

    gateway_connect_parser = gateway_subparsers.add_parser("connect", help="Add a wallet for a chain")
    gateway_connect_parser.add_argument("chain", nargs="?", default=None, help="Blockchain chain (e.g., ethereum, solana)")
    gateway_connect_parser.set_defaults(func=hummingbot.gateway_connect)

    gateway_cert_parser = gateway_subparsers.add_parser("generate-certs", help="Create SSL certificate")
    gateway_cert_parser.set_defaults(func=hummingbot.generate_certs)

    gateway_list_parser = gateway_subparsers.add_parser("list", help="List available connectors")
    gateway_list_parser.set_defaults(func=hummingbot.gateway_list)

    gateway_lp_parser = gateway_subparsers.add_parser("lp", help="Manage liquidity positions")
    gateway_lp_parser.add_argument("connector", nargs="?", type=str, help="Connector name/type (e.g., raydium/amm)")
    gateway_lp_parser.add_argument("action", nargs="?", type=str, choices=["add-liquidity", "remove-liquidity", "position-info", "collect-fees"], help="LP action to perform")
    gateway_lp_parser.add_argument("trading_pair", nargs="?", default=None, help="Trading pair (e.g., WETH-USDC)")
    gateway_lp_parser.set_defaults(func=hummingbot.gateway_lp)

    gateway_ping_parser = gateway_subparsers.add_parser("ping", help="Test node and chain/network status")
    gateway_ping_parser.add_argument("chain", nargs="?", default=None, help="Specific chain to test (optional)")
    gateway_ping_parser.set_defaults(func=hummingbot.gateway_ping)

    gateway_pool_parser = gateway_subparsers.add_parser("pool", help="View or update pool information")
    gateway_pool_parser.add_argument("connector", nargs="?", default=None, help="Connector name/type (e.g., uniswap/amm)")
    gateway_pool_parser.add_argument("trading_pair", nargs="?", default=None, help="Trading pair (e.g., ETH-USDC)")
    gateway_pool_parser.add_argument("action", nargs="?", default=None, help="Action to perform (update)")
    gateway_pool_parser.add_argument("args", nargs="*", help="Additional arguments: <address> for direct pool update")
    gateway_pool_parser.set_defaults(func=hummingbot.gateway_pool)

    gateway_swap_parser = gateway_subparsers.add_parser(
        "swap",
        help="Swap tokens")
    gateway_swap_parser.add_argument("connector", nargs="?", default=None,
                                     help="Connector name/type (e.g., jupiter/router)")
    gateway_swap_parser.add_argument("args", nargs="*",
                                     help="Arguments: [base-quote] [side] [amount]. "
                                          "Interactive mode if not all provided. "
                                          "Example: gateway swap uniswap ETH-USDC BUY 0.1")
    gateway_swap_parser.set_defaults(func=hummingbot.gateway_swap)

    gateway_token_parser = gateway_subparsers.add_parser("token", help="View or update token information")
    gateway_token_parser.add_argument("symbol_or_address", nargs="?", default=None, help="Token symbol or address")
    gateway_token_parser.add_argument("action", nargs="?", default=None, help="Action to perform (update)")
    gateway_token_parser.set_defaults(func=hummingbot.gateway_token)

    exit_parser = subparsers.add_parser("exit", help="Exit and cancel all outstanding orders")
    exit_parser.add_argument("-f", "--force", action="store_true", help="Force exit without canceling outstanding orders",
                             default=False)
    exit_parser.set_defaults(func=hummingbot.exit)

    export_parser = subparsers.add_parser("export", help="Export secure information")
    export_parser.add_argument("option", nargs="?", choices=("keys", "trades"), help="Export choices")
    export_parser.set_defaults(func=hummingbot.export)

    ticker_parser = subparsers.add_parser("ticker", help="Show market ticker of current order book")
    ticker_parser.add_argument("--live", default=False, action="store_true", dest="live", help="Show ticker updates")
    ticker_parser.add_argument("--exchange", type=str, dest="exchange", help="The exchange of the market")
    ticker_parser.add_argument("--market", type=str, dest="market", help="The market (trading pair) of the order book")
    ticker_parser.set_defaults(func=hummingbot.ticker)

    mqtt_parser = subparsers.add_parser("mqtt", help="Manage the MQTT broker bridge")
    mqtt_subparsers = mqtt_parser.add_subparsers()
    mqtt_start_parser = mqtt_subparsers.add_parser("start", help="Start the MQTT broker bridge")
    mqtt_start_parser.add_argument(
        "-t",
        "--timeout",
        default=30.0,
        type=float,
        dest="timeout",
        help="Bridge connection timeout"
    )
    mqtt_start_parser.set_defaults(func=hummingbot.mqtt_start)
    mqtt_stop_parser = mqtt_subparsers.add_parser("stop", help="Stop the MQTT Bridge")
    mqtt_stop_parser.set_defaults(func=hummingbot.mqtt_stop)
    mqtt_restart_parser = mqtt_subparsers.add_parser("restart", help="Restart the MQTT Bridge")
    mqtt_restart_parser.add_argument(
        "-t",
        "--timeout",
        default=30.0,
        type=float,
        dest="timeout",
        help="Bridge connection timeout"
    )
    mqtt_restart_parser.set_defaults(func=hummingbot.mqtt_restart)

    rate_parser = subparsers.add_parser('rate', help="Show rate of a given trading pair")
    rate_parser.add_argument("-p", "--pair", default=None,
                             dest="pair", help="The market trading pair for which you want to get a rate.")
    rate_parser.add_argument("-t", "--token", default=None,
                             dest="token", help="The token who's value you want to get.")
    rate_parser.set_defaults(func=hummingbot.rate)

    for name, command_tab in command_tabs.items():
        o_parser = subparsers.add_parser(name, help=command_tab.tab_class.get_command_help_message())
        for arg_name, arg_properties in command_tab.tab_class.get_command_arguments().items():
            o_parser.add_argument(arg_name, **arg_properties)
        o_parser.add_argument("-c", "--close", default=False, action="store_true", dest="close",
                              help=f"To close the {name} tab.")

    return parser



import asyncio
from collections import defaultdict
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional

import hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_constants as CONSTANTS
import hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_web_utils as web_utils
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_derivative import AevoPerpetualDerivative


class AevoPerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):
    _bpobds_logger: Optional[HummingbotLogger] = None
    _trading_pair_symbol_map: Dict[str, Mapping[str, str]] = {}
    _mapping_initialization_lock = asyncio.Lock()

    def __init__(
            self,
            trading_pairs: List[str],
            connector: 'AevoPerpetualDerivative',
            api_factory: WebAssistantsFactory,
            domain: str = CONSTANTS.DEFAULT_DOMAIN,
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._domain = domain
        self._trading_pairs: List[str] = trading_pairs
        self._message_queue: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._snapshot_messages_queue_key = "order_book_snapshot"

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        funding = await self._connector._api_get(
            path_url=CONSTANTS.FUNDING_PATH_URL,
            params={"instrument_name": ex_trading_pair},
        )
        instrument = await self._connector._api_get(
            path_url=f"{CONSTANTS.INSTRUMENT_PATH_URL}/{ex_trading_pair}",
            limit_id=CONSTANTS.INSTRUMENT_PATH_URL,
        )
        next_epoch_ns = int(funding.get("next_epoch", "0"))
        funding_info = FundingInfo(
            trading_pair=trading_pair,
            index_price=Decimal(instrument.get("index_price", "0")),
            mark_price=Decimal(instrument.get("mark_price", "0")),
            next_funding_utc_timestamp=int(next_epoch_ns * 1e-9),
            rate=Decimal(funding.get("funding_rate", "0")),
        )
        return funding_info

    async def listen_for_funding_info(self, output: asyncio.Queue):
        while True:
            try:
                for trading_pair in self._trading_pairs:
                    funding_info = await self.get_funding_info(trading_pair)
                    funding_info_update = FundingInfoUpdate(
                        trading_pair=trading_pair,
                        index_price=funding_info.index_price,
                        mark_price=funding_info.mark_price,
                        next_funding_utc_timestamp=funding_info.next_funding_utc_timestamp,
                        rate=funding_info.rate,
                    )
                    output.put_nowait(funding_info_update)
                await self._sleep(CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error when processing public funding info updates from exchange")
                await self._sleep(CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        data = await self._connector._api_get(
            path_url=CONSTANTS.ORDERBOOK_PATH_URL,
            params={"instrument_name": ex_trading_pair},
        )
        return data

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        timestamp = int(snapshot_response["last_updated"]) * 1e-9
        snapshot_msg: OrderBookMessage = OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": trading_pair,
            "update_id": int(snapshot_response["last_updated"]),
            "bids": [[float(i[0]), float(i[1])] for i in snapshot_response.get("bids", [])],
            "asks": [[float(i[0]), float(i[1])] for i in snapshot_response.get("asks", [])],
        }, timestamp=timestamp)
        return snapshot_msg

    async def _connected_websocket_assistant(self) -> WSAssistant:
        url = f"{web_utils.wss_url(self._domain)}"
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)

        return ws

    async def _subscribe_channels(self, ws: WSAssistant):
        try:
            for trading_pair in self._trading_pairs:
                symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

                trades_payload = {
                    "op": "subscribe",
                    "data": [f"{CONSTANTS.WS_TRADE_CHANNEL}:{symbol}"],
                }
                subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trades_payload)

                order_book_payload = {
                    "op": "subscribe",
                    "data": [f"{CONSTANTS.WS_ORDERBOOK_CHANNEL}:{symbol}"],
                }
                subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

                await ws.send(subscribe_trade_request)
                await ws.send(subscribe_orderbook_request)

                self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error("Unexpected error occurred subscribing to order book data streams.")
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "channel" in event_message:
            stream_name = event_message.get("channel")
            if stream_name.startswith(f"{CONSTANTS.WS_ORDERBOOK_CHANNEL}:"):
                msg_type = event_message.get("data", {}).get("type")
                if msg_type == "snapshot":
                    channel = self._snapshot_messages_queue_key
                else:
                    channel = self._diff_messages_queue_key
            elif stream_name.startswith(f"{CONSTANTS.WS_TRADE_CHANNEL}:"):
                channel = self._trade_messages_queue_key
            else:
                self.logger().warning(f"Unknown WS channel received: {stream_name}")
        return channel

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        data = raw_message["data"]
        timestamp = int(data["last_updated"]) * 1e-9
        instrument_name = raw_message["data"]["instrument_name"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(instrument_name)
        order_book_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.DIFF, {
            "trading_pair": trading_pair,
            "update_id": int(data["last_updated"]),
            "bids": [[float(i[0]), float(i[1])] for i in data.get("bids", [])],
            "asks": [[float(i[0]), float(i[1])] for i in data.get("asks", [])],
        }, timestamp=timestamp)
        message_queue.put_nowait(order_book_message)

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        data = raw_message["data"]
        timestamp = int(data["last_updated"]) * 1e-9
        instrument_name = raw_message["data"]["instrument_name"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(instrument_name)
        order_book_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": trading_pair,
            "update_id": int(data["last_updated"]),
            "bids": [[float(i[0]), float(i[1])] for i in data.get("bids", [])],
            "asks": [[float(i[0]), float(i[1])] for i in data.get("asks", [])],
        }, timestamp=timestamp)
        message_queue.put_nowait(order_book_message)

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        data = raw_message["data"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(
            data["instrument_name"])
        timestamp = int(data.get("created_timestamp", "0")) * 1e-9
        trade_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.TRADE, {
            "trading_pair": trading_pair,
            "trade_type": float(TradeType.BUY.value) if data["side"] == "buy" else float(TradeType.SELL.value),
            "trade_id": str(data["trade_id"]),
            "price": float(data["price"]),
            "amount": float(data["amount"]),
        }, timestamp=timestamp)
        message_queue.put_nowait(trade_message)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        pass

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot subscribe to {trading_pair}: WebSocket connection not established."
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            trades_payload = {
                "op": "subscribe",
                "data": [f"{CONSTANTS.WS_TRADE_CHANNEL}:{symbol}"],
            }
            order_book_payload = {
                "op": "subscribe",
                "data": [f"{CONSTANTS.WS_ORDERBOOK_CHANNEL}:{symbol}"],
            }

            await self._ws_assistant.send(WSJSONRequest(payload=trades_payload))
            await self._ws_assistant.send(WSJSONRequest(payload=order_book_payload))

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Successfully subscribed to {trading_pair}")
            return True
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().error(f"Error subscribing to {trading_pair}: {e}")
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        if self._ws_assistant is None:
            self.logger().warning(
                f"Cannot unsubscribe from {trading_pair}: WebSocket connection not established."
            )
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

            trades_payload = {
                "op": "unsubscribe",
                "data": [f"{CONSTANTS.WS_TRADE_CHANNEL}:{symbol}"],
            }
            order_book_payload = {
                "op": "unsubscribe",
                "data": [f"{CONSTANTS.WS_ORDERBOOK_CHANNEL}:{symbol}"],
            }

            await self._ws_assistant.send(WSJSONRequest(payload=trades_payload))
            await self._ws_assistant.send(WSJSONRequest(payload=order_book_payload))

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Successfully unsubscribed from {trading_pair}")
            return True
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().error(f"Error unsubscribing from {trading_pair}: {e}")
            return False



import typing

from hummingbot.connector.budget_checker import BudgetChecker
from hummingbot.connector.perpetual_trading import PerpetualTrading
from hummingbot.core.data_type.order_candidate import PerpetualOrderCandidate

if typing.TYPE_CHECKING:  # avoid circular import problems
    from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase


class PerpetualBudgetChecker(BudgetChecker):
    def __init__(self, exchange: "PerpetualDerivativePyBase"):
        """
        In the case of derived instruments, the collateral can be any token.
        To get this information, this class uses the `get_buy_collateral_token`
        and `get_sell_collateral_token` methods provided by the `PerpetualTrading` interface.
        """
        super().__init__(exchange)
        self._validate_perpetual_connector()

    def _validate_perpetual_connector(self):
        from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
        if not isinstance(self._exchange, (PerpetualTrading, PerpetualDerivativePyBase)):
            raise TypeError(
                f"{self.__class__} must be passed an exchange implementing the {PerpetualTrading} interface."
            )

    def _lock_available_collateral(self, order_candidate: PerpetualOrderCandidate):
        if not order_candidate.position_close:
            super()._lock_available_collateral(order_candidate)



cdef class TradingRule:
    cdef:
        public str trading_pair
        public object min_order_size                   # Calculated min base asset size based on last trade price
        public object max_order_size                   # Calculated max base asset size
        public object min_price_increment              # Min tick size difference accepted (e.g. 0.1)
        public object min_base_amount_increment        # Min step size of base asset amount (e.g. 0.01)
        public object min_quote_amount_increment       # Min step size of quote asset amount (e.g. 0.01)
        public object max_price_significant_digits     # Max # of significant digits in a price
        public object min_notional_size                # Notional value = price * quantity, min accepted (e.g. 3.001)
        public object min_order_value                  # Calculated min base asset value based on the minimum accepted trade value (e.g. 0.078LTC is ~50,000 Satoshis)
        public bint supports_limit_orders              # if limit order is allowed for this trading pair
        public bint supports_market_orders             # if market order is allowed for this trading pair
        public object buy_order_collateral_token       # Indicates the collateral token used for buy orders
        public object sell_order_collateral_token      # Indicates the collateral token used for sell orders



import asyncio
import logging
import time
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional

from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger


class OrderBookTrackerDataSource(metaclass=ABCMeta):
    FULL_ORDER_BOOK_RESET_DELTA_SECONDS = 60 * 60

    _logger: Optional[HummingbotLogger] = None

    def __init__(self, trading_pairs: List[str]):
        self._trade_messages_queue_key = "trade"
        self._diff_messages_queue_key = "order_book_diff"
        self._snapshot_messages_queue_key = "order_book_snapshot"

        self._trading_pairs: List[str] = trading_pairs
        self._order_book_create_function = lambda: OrderBook()
        self._message_queue: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._ws_assistant: Optional[WSAssistant] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(HummingbotLogger.logger_name_for_class(cls))
        return cls._logger

    @property
    def order_book_create_function(self) -> Callable[[], OrderBook]:
        return self._order_book_create_function

    @order_book_create_function.setter
    def order_book_create_function(self, func: Callable[[], OrderBook]):
        self._order_book_create_function = func

    @abstractmethod
    async def get_last_traded_prices(self, trading_pairs: List[str], domain: Optional[str] = None) -> Dict[str, float]:
        """
        Return a dictionary the trading_pair as key and the current price as value for each trading pair passed as
        parameter.
        This method is required by the order book tracker, to get the last traded prices when no new public trades
        are notified by the exchange.

        :param trading_pairs: list of trading pairs to get the prices for
        :param domain: which domain we are connecting to

        :return: Dictionary of associations between token pair and its latest price
        """
        raise NotImplementedError

    async def get_new_order_book(self, trading_pair: str) -> OrderBook:
        """
        Creates a local instance of the exchange order book for a particular trading pair

        :param trading_pair: the trading pair for which the order book has to be retrieved

        :return: a local copy of the current order book in the exchange
        """
        snapshot_msg: OrderBookMessage = await self._order_book_snapshot(trading_pair=trading_pair)
        order_book: OrderBook = self.order_book_create_function()
        order_book.apply_snapshot(snapshot_msg.bids, snapshot_msg.asks, snapshot_msg.update_id)
        return order_book

    async def listen_for_subscriptions(self):
        """
        Connects to the trade events and order diffs websocket endpoints and listens to the messages sent by the
        exchange. Each message is stored in its own queue.
        """
        ws: Optional[WSAssistant] = None
        while True:
            try:
                ws: WSAssistant = await self._connected_websocket_assistant()
                self._ws_assistant = ws  # Store reference for incremental subscriptions
                await self._subscribe_channels(ws)
                await self._process_websocket_messages(websocket_assistant=ws)
            except asyncio.CancelledError:
                raise
            except ConnectionError as connection_exception:
                self.logger().warning(f"The websocket connection was closed ({connection_exception})")
            except Exception:
                self.logger().exception(
                    "Unexpected error occurred when listening to order book streams. Retrying in 5 seconds...",
                )
                await self._sleep(1.0)
            finally:
                self._ws_assistant = None  # Clear reference on disconnect
                await self._on_order_stream_interruption(websocket_assistant=ws)

    async def listen_for_order_book_diffs(self, ev_loop: asyncio.AbstractEventLoop, output: asyncio.Queue):
        """
        Reads the order diffs events queue. For each event creates a diff message instance and adds it to the
        output queue

        :param ev_loop: the event loop the method will run in
        :param output: a queue to add the created diff messages
        """
        message_queue = self._message_queue[self._diff_messages_queue_key]
        while True:
            try:
                diff_event = await message_queue.get()
                await self._parse_order_book_diff_message(raw_message=diff_event, message_queue=output)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error when processing public order book updates from exchange")

    async def listen_for_order_book_snapshots(self, ev_loop: asyncio.AbstractEventLoop, output: asyncio.Queue):
        """
        Reads the order snapshot events queue. For each event it creates a snapshot message instance and adds it to the
        output queue.
        This method also request the full order book content from the exchange using HTTP requests if it does not
        receive events during one hour.

        :param ev_loop: the event loop the method will run in
        :param output: a queue to add the created snapshot messages
        """
        message_queue = self._message_queue[self._snapshot_messages_queue_key]
        while True:
            try:
                try:
                    snapshot_event = await asyncio.wait_for(message_queue.get(),
                                                            timeout=self.FULL_ORDER_BOOK_RESET_DELTA_SECONDS)
                    await self._parse_order_book_snapshot_message(raw_message=snapshot_event, message_queue=output)
                except asyncio.TimeoutError:
                    await self._request_order_book_snapshots(output=output)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error when processing public order book snapshots from exchange")
                await self._sleep(1.0)

    async def listen_for_trades(self, ev_loop: asyncio.AbstractEventLoop, output: asyncio.Queue):
        """
        Reads the trade events queue. For each event creates a trade message instance and adds it to the output queue

        :param ev_loop: the event loop the method will run in
        :param output: a queue to add the created trade messages
        """
        message_queue = self._message_queue[self._trade_messages_queue_key]
        while True:
            try:
                trade_event = await message_queue.get()
                await self._parse_trade_message(raw_message=trade_event, message_queue=output)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error when processing public trade updates from exchange")

    async def _request_order_book_snapshots(self, output: asyncio.Queue):
        for trading_pair in self._trading_pairs:
            try:
                snapshot = await self._order_book_snapshot(trading_pair=trading_pair)
                output.put_nowait(snapshot)
            except Exception:
                self.logger().exception(f"Unexpected error fetching order book snapshot for {trading_pair}.")
                raise

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        """
        Create an instance of OrderBookMessage of type OrderBookMessageType.TRADE

        :param raw_message: the JSON dictionary of the public trade event
        :param message_queue: queue where the parsed messages should be stored in
        """
        raise NotImplementedError

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        """
        Create an instance of OrderBookMessage of type OrderBookMessageType.DIFF

        :param raw_message: the JSON dictionary of the public trade event
        :param message_queue: queue where the parsed messages should be stored in
        """
        raise NotImplementedError

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        """
        Create an instance of OrderBookMessage of type OrderBookMessageType.SNAPSHOT

        :param raw_message: the JSON dictionary of the public trade event
        :param message_queue: queue where the parsed messages should be stored in
        """
        raise NotImplementedError

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        raise NotImplementedError

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange

        :return: an instance of WSAssistant connected to the exchange
        """
        raise NotImplementedError

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param ws: the websocket assistant used to connect to the exchange
        """
        raise NotImplementedError

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        """
        Identifies the channel for a particular event message. Used to find the correct queue to add the message in

        :param event_message: the event received through the websocket connection

        :return: the message channel
        """
        raise NotImplementedError

    async def _process_message_for_unknown_channel(
        self, event_message: Dict[str, Any], websocket_assistant: WSAssistant
    ):
        """
        Processes a message coming from a not identified channel.
        Does nothing by default but allows subclasses to reimplement

        :param event_message: the event received through the websocket connection
        :param websocket_assistant: the websocket connection to use to interact with the exchange
        """
        pass

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant):
        async for ws_response in websocket_assistant.iter_messages():
            data: Dict[str, Any] = ws_response.data
            if data is not None:  # data will be None when the websocket is disconnected
                channel: str = self._channel_originating_message(event_message=data)
                valid_channels = self._get_messages_queue_keys()
                if channel in valid_channels:
                    self._message_queue[channel].put_nowait(data)
                else:
                    await self._process_message_for_unknown_channel(
                        event_message=data, websocket_assistant=websocket_assistant
                    )

    def _get_messages_queue_keys(self) -> List[str]:
        return [self._snapshot_messages_queue_key, self._diff_messages_queue_key, self._trade_messages_queue_key]

    async def _on_order_stream_interruption(self, websocket_assistant: Optional[WSAssistant] = None):
        websocket_assistant and await websocket_assistant.disconnect()

    async def _sleep(self, delay):
        """
        Function added only to facilitate patching the sleep in unit tests without affecting the asyncio module
        """
        await asyncio.sleep(delay)

    def _time(self):
        return time.time()

    def add_trading_pair(self, trading_pair: str):
        """
        Adds a trading pair to the internal list of tracked pairs.

        :param trading_pair: the trading pair to add
        """
        if trading_pair not in self._trading_pairs:
            self._trading_pairs.append(trading_pair)

    def remove_trading_pair(self, trading_pair: str):
        """
        Removes a trading pair from the internal list of tracked pairs.

        :param trading_pair: the trading pair to remove
        """
        if trading_pair in self._trading_pairs:
            self._trading_pairs.remove(trading_pair)

    @abstractmethod
    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribes to order book and trade channels for a single trading pair on an
        existing WebSocket connection.

        :param trading_pair: the trading pair to subscribe to
        :return: True if subscription was successful, False otherwise
        """
        raise NotImplementedError

    @abstractmethod
    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribes from order book and trade channels for a single trading pair on an
        existing WebSocket connection.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if unsubscription was successful, False otherwise
        """
        raise NotImplementedError



# A single source of truth for constant variables related to the exchange

from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "bitmart"
REST_URL = "https://api-cloud.bitmart.com"
WSS_PUBLIC_URL = "wss://ws-manager-compress.bitmart.com/api?protocol=1.1"
WSS_PRIVATE_URL = "wss://ws-manager-compress.bitmart.com/user?protocol=1.1"
WS_PING_TIMEOUT = 20 * 0.8

DEFAULT_DOMAIN = ""
MAX_ORDER_ID_LEN = 32
HBOT_ORDER_ID_PREFIX = ""
BROKER_ID = "hummingbotfound"

PUBLIC_TRADE_CHANNEL_NAME = "spot/trade"
PUBLIC_DEPTH_CHANNEL_NAME = "spot/depth50"
PRIVATE_ORDER_PROGRESS_CHANNEL_NAME = "spot/user/order"

# REST API ENDPOINTS
CHECK_NETWORK_PATH_URL = "system/service"
GET_TRADING_RULES_PATH_URL = "spot/v1/symbols/details"
GET_LAST_TRADING_PRICES_PATH_URL = "spot/quotation/v3/ticker"
GET_ORDER_BOOK_PATH_URL = "spot/quotation/v3/books"
CREATE_ORDER_PATH_URL = "spot/v2/submit_order"
CANCEL_ORDER_PATH_URL = "spot/v3/cancel_order"
GET_ACCOUNT_SUMMARY_PATH_URL = "spot/v1/wallet"
GET_ORDER_DETAIL_PATH_URL = "spot/v4/query/order"
GET_TRADE_DETAIL_PATH_URL = "spot/v4/query/order-trades"
SERVER_TIME_PATH = "system/time"

# WS API ENDPOINTS
WS_CONNECT = "WSConnect"
WS_SUBSCRIBE = "WSSubscribe"

# BitMart has a per method API limit
RATE_LIMITS = [
    RateLimit(limit_id=CHECK_NETWORK_PATH_URL, limit=10, time_interval=1),
    RateLimit(limit_id=GET_TRADING_RULES_PATH_URL, limit=30, time_interval=5),
    RateLimit(limit_id=GET_LAST_TRADING_PRICES_PATH_URL, limit=30, time_interval=5),
    RateLimit(limit_id=GET_ORDER_BOOK_PATH_URL, limit=30, time_interval=5),
    RateLimit(limit_id=CREATE_ORDER_PATH_URL, limit=150, time_interval=5),
    RateLimit(limit_id=CANCEL_ORDER_PATH_URL, limit=150, time_interval=5),
    RateLimit(limit_id=GET_ACCOUNT_SUMMARY_PATH_URL, limit=30, time_interval=5),
    RateLimit(limit_id=GET_ORDER_DETAIL_PATH_URL, limit=150, time_interval=5),
    RateLimit(limit_id=GET_TRADE_DETAIL_PATH_URL, limit=30, time_interval=5),
    RateLimit(limit_id=SERVER_TIME_PATH, limit=10, time_interval=1),
    RateLimit(limit_id=WS_CONNECT, limit=30, time_interval=60),
    RateLimit(limit_id=WS_SUBSCRIBE, limit=100, time_interval=10),
]

ORDER_STATE = {
    "failed": OrderState.FAILED,
    "new": OrderState.OPEN,
    "partially_filled": OrderState.PARTIALLY_FILLED,
    "filled": OrderState.FILLED,
    "partially_canceled": OrderState.CANCELED,
    "canceled": OrderState.CANCELED
}



from enum import Enum
from typing import Dict, Optional

from hummingbot.connector.exchange.foxbit import foxbit_constants as CONSTANTS
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType


class FoxbitTradeFields(Enum):
    ID = 0
    INSTRUMENTID = 1
    QUANTITY = 2
    PRICE = 3
    ORDERMAKERID = 4
    ORDERTAKERID = 5
    CREATEDAT = 6
    TREND = 7
    SIDE = 8
    FIXED_BOOL = 9
    FIXED_INT = 10


class FoxbitOrderBookFields(Enum):
    MDUPDATEID = 0
    ACCOUNTS = 1
    ACTIONDATETIME = 2
    ACTIONTYPE = 3
    LASTTRADEPRICE = 4
    ORDERS = 5
    PRICE = 6
    PRODUCTPAIRCODE = 7
    QUANTITY = 8
    SIDE = 9


class FoxbitOrderBookAction(Enum):
    NEW = 0
    UPDATE = 1
    DELETION = 2


class FoxbitOrderBookSide(Enum):
    BID = 0
    ASK = 1


class FoxbitOrderBookItem(Enum):
    PRICE = 0
    QUANTITY = 1


class FoxbitOrderBook(OrderBook):
    _bids = {}
    _asks = {}

    @classmethod
    def trade_message_from_exchange(cls,
                                    msg: Dict[str, any],
                                    metadata: Optional[Dict] = None,
                                    ):
        """
        Creates a trade message with the information from the trade event sent by the exchange
        :param msg: the trade event details sent by the exchange
        :param metadata: a dictionary with extra information to add to trade message
        :return: a trade message with the details of the trade as provided by the exchange
        """
        ts = int(msg[FoxbitTradeFields.CREATEDAT.value])
        return OrderBookMessage(OrderBookMessageType.TRADE, {
            "trading_pair": metadata["trading_pair"],
            "trade_type": float(TradeType.SELL.value) if msg[FoxbitTradeFields.SIDE.value] == 1 else float(TradeType.BUY.value),
            "trade_id": msg[FoxbitTradeFields.ID.value],
            "update_id": ts,
            "price": '%.10f' % float(msg[FoxbitTradeFields.PRICE.value]),
            "amount": '%.10f' % float(msg[FoxbitTradeFields.QUANTITY.value])
        }, timestamp=ts * 1e-3)

    @classmethod
    def snapshot_message_from_exchange(cls,
                                       msg: Dict[str, any],
                                       timestamp: float,
                                       metadata: Optional[Dict] = None,
                                       ) -> OrderBookMessage:
        """
        Creates a snapshot message with the order book snapshot message
        :param msg: the response from the exchange when requesting the order book snapshot
        :param timestamp: the snapshot timestamp
        :param metadata: a dictionary with extra information to add to the snapshot data
        :return: a snapshot message with the snapshot information received from the exchange

        sample of msg {'sequence_id': 5972127, 'asks': [['140999.9798', '0.00007093'], ['140999.9899', '0.10646516'], ['140999.99', '0.01166287'], ['141000.0', '0.00024751'], ['141049.9999', '0.3688'], ['141050.0', '0.00184094'], ['141099.0', '0.00007087'], ['141252.9994', '0.02374105'], ['141253.0', '0.5786'], ['141275.0', '0.00707839'], ['141299.0', '0.00007077'], ['141317.9492', '0.814357'], ['141323.9741', '0.0039086'], ['141339.358', '0.64833964']], 'bids': [[['140791.4571', '0.0000569'], ['140791.4471', '0.00000028'], ['140791.4371', '0.0000289'], ['140791.4271', '0.00018672'], ['140512.4635', '0.06396371'], ['140512.4632', '0.3688'], ['140506.0', '0.5786'], ['140499.5014', '0.1'], ['140377.2678', '0.00976774'], ['140300.0', '0.005866'], ['140054.3859', '0.14746'], ['140054.1159', '3.45282018'], ['140032.8321', '1.2267452'], ['140025.553', '1.12483605']]}
        """
        cls.logger().info(f'Refreshing order book to {metadata["trading_pair"]}.')

        cls._bids = {}
        cls._asks = {}

        for item in msg["bids"]:
            cls.update_order_book('%.10f' % float(item[FoxbitOrderBookItem.QUANTITY.value]),
                                  '%.10f' % float(item[FoxbitOrderBookItem.PRICE.value]),
                                  FoxbitOrderBookSide.BID)

        for item in msg["asks"]:
            cls.update_order_book('%.10f' % float(item[FoxbitOrderBookItem.QUANTITY.value]),
                                  '%.10f' % float(item[FoxbitOrderBookItem.PRICE.value]),
                                  FoxbitOrderBookSide.ASK)

        return OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": metadata["trading_pair"],
            "update_id": int(msg["sequence_id"]),
            "bids": [[price, quantity] for price, quantity in cls._bids.items()],
            "asks": [[price, quantity] for price, quantity in cls._asks.items()]
        }, timestamp=timestamp)

    @classmethod
    def diff_message_from_exchange(cls,
                                   msg: Dict[str, any],
                                   timestamp: Optional[float] = None,
                                   metadata: Optional[Dict] = None,
                                   ) -> OrderBookMessage:
        """
        Creates a diff message with the changes in the order book received from the exchange
        :param msg: the changes in the order book
        :param timestamp: the timestamp of the difference
        :param metadata: a dictionary with extra information to add to the difference data
        :return: a diff message with the changes in the order book notified by the exchange

        sample of msg = [5971940, 0, 1683735920192, 2, 140999.9798, 0, 140688.6227, 1, 0, 0]
        """
        trading_pair = metadata["trading_pair"]
        order_book_id = int(msg[FoxbitOrderBookFields.MDUPDATEID.value])
        prc = '%.10f' % float(msg[FoxbitOrderBookFields.PRICE.value])
        qty = '%.10f' % float(msg[FoxbitOrderBookFields.QUANTITY.value])

        if msg[FoxbitOrderBookFields.ACTIONTYPE.value] == FoxbitOrderBookAction.DELETION.value:
            qty = '0'

        if msg[FoxbitOrderBookFields.SIDE.value] == FoxbitOrderBookSide.BID.value:

            return OrderBookMessage(
                OrderBookMessageType.DIFF, {
                    "trading_pair": trading_pair,
                    "update_id": order_book_id,
                    "bids": [[prc, qty]],
                    "asks": [],
                }, timestamp=int(msg[FoxbitOrderBookFields.ACTIONDATETIME.value]))

        if msg[FoxbitOrderBookFields.SIDE.value] == FoxbitOrderBookSide.ASK.value:
            return OrderBookMessage(
                OrderBookMessageType.DIFF, {
                    "trading_pair": trading_pair,
                    "update_id": order_book_id,
                    "bids": [],
                    "asks": [[prc, qty]],
                }, timestamp=int(msg[FoxbitOrderBookFields.ACTIONDATETIME.value]))

    @classmethod
    def update_order_book(cls, quantity: str, price: str, side: FoxbitOrderBookSide):
        q = float(quantity)
        p = float(price)

        if side == FoxbitOrderBookSide.BID:
            cls._bids[p] = q
            if len(cls._bids) > CONSTANTS.ORDER_BOOK_DEPTH:
                min_bid = min(cls._bids.keys())
                del cls._bids[min_bid]

            cls._bids = dict(sorted(cls._bids.items(), reverse=True))
            return

        if side == FoxbitOrderBookSide.ASK:
            cls._asks[p] = q
            if len(cls._asks) > CONSTANTS.ORDER_BOOK_DEPTH:
                max_ask = max(cls._asks.keys())
                del cls._asks[max_ask]

            cls._asks = dict(sorted(cls._asks.items()))
            return



"""
XRPL Transaction Pipeline

Serializes all XRPL transaction submissions to prevent sequence number race conditions.

Architecture:
- Single FIFO queue for all transaction submissions
- Pipeline loop processes one transaction at a time
- Configurable delay between submissions
- Since only one transaction is processed at a time, autofill always gets the correct sequence

This pipeline is shared across all wallet-specific transaction pools to ensure
global serialization of transaction submissions.
"""

import asyncio
import logging
import time
import uuid
from typing import Any, Awaitable, Optional, Tuple

from hummingbot.connector.exchange.xrpl import xrpl_constants as CONSTANTS
from hummingbot.connector.exchange.xrpl.xrpl_utils import XRPLSystemBusyError
from hummingbot.logger import HummingbotLogger


class XRPLTransactionPipeline:
    """
    Serialized transaction submission pipeline for XRPL.

    All transaction submissions go through this pipeline to ensure:
    1. Only one transaction is processed at a time
    2. Proper spacing between submissions
    3. Sequence numbers are correctly assigned by autofill

    This prevents race conditions where multiple concurrent autofills
    could get the same sequence number.
    """
    _logger: Optional[HummingbotLogger] = None

    def __init__(
        self,
        max_queue_size: int = CONSTANTS.PIPELINE_MAX_QUEUE_SIZE,
        submission_delay_ms: int = CONSTANTS.PIPELINE_SUBMISSION_DELAY_MS,
    ):
        """
        Initialize the transaction pipeline.

        Args:
            max_queue_size: Maximum pending submissions in the queue
            submission_delay_ms: Delay in milliseconds between submissions
        """
        self._max_queue_size = max_queue_size
        self._delay_seconds = submission_delay_ms / 1000.0

        # FIFO queue: (coroutine, future, submission_id)
        self._submission_queue: asyncio.Queue[Tuple[Awaitable, asyncio.Future, str]] = asyncio.Queue(
            maxsize=max_queue_size
        )
        self._pipeline_task: Optional[asyncio.Task] = None
        self._running = False
        self._started = False  # For lazy initialization

        # Statistics
        self._submissions_processed = 0
        self._submissions_failed = 0
        self._total_latency_ms = 0.0

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(HummingbotLogger.logger_name_for_class(cls))
        return cls._logger

    @property
    def is_running(self) -> bool:
        """Check if the pipeline is running."""
        return self._running

    @property
    def queue_size(self) -> int:
        """Get the current queue size."""
        return self._submission_queue.qsize()

    @property
    def stats(self) -> dict:
        """Get pipeline statistics."""
        total = self._submissions_processed + self._submissions_failed
        avg_latency = self._total_latency_ms / total if total > 0 else 0.0
        return {
            "queue_size": self.queue_size,
            "submissions_processed": self._submissions_processed,
            "submissions_failed": self._submissions_failed,
            "avg_latency_ms": round(avg_latency, 2),
        }

    async def start(self):
        """Start the pipeline loop."""
        if self._running:
            self.logger().warning("[PIPELINE] Pipeline is already running")
            return

        self._running = True
        self._started = True
        self._pipeline_task = asyncio.create_task(self._pipeline_loop())

        self.logger().debug(
            f"[PIPELINE] Started with {self._delay_seconds * 1000:.0f}ms delay between submissions"
        )

    async def stop(self):
        """Stop the pipeline and cancel pending submissions."""
        if not self._running:
            return

        self._running = False
        self.logger().debug("[PIPELINE] Stopping...")

        # Cancel pipeline task
        if self._pipeline_task is not None:
            self._pipeline_task.cancel()
            try:
                await self._pipeline_task
            except asyncio.CancelledError:
                pass
            self._pipeline_task = None

        # Cancel pending submissions
        cancelled_count = 0
        while not self._submission_queue.empty():
            try:
                _, future, submission_id = self._submission_queue.get_nowait()
                if not future.done():
                    future.cancel()
                    cancelled_count += 1
                    self.logger().debug(f"[PIPELINE] Cancelled pending submission {submission_id}")
            except asyncio.QueueEmpty:
                break

        self.logger().debug(
            f"[PIPELINE] Stopped, cancelled {cancelled_count} pending submissions"
        )

    async def _ensure_started(self):
        """Ensure the pipeline is started (lazy initialization)."""
        if not self._started:
            await self.start()

    async def submit(
        self,
        coro: Awaitable,
        submission_id: Optional[str] = None,
    ) -> Any:
        """
        Submit a coroutine to the serialized pipeline.

        All XRPL transaction submissions should go through this method
        to ensure they are processed one at a time.

        Args:
            coro: The coroutine to execute (typically autofill/sign/submit)
            submission_id: Optional identifier for tracing

        Returns:
            The result from the coroutine

        Raises:
            XRPLSystemBusyError: If the pipeline queue is full
            Exception: Any exception raised by the coroutine
        """
        # Lazy start
        await self._ensure_started()

        if not self._running:
            raise XRPLSystemBusyError("Pipeline is not running")

        # Generate submission_id if not provided
        if submission_id is None:
            submission_id = str(uuid.uuid4())[:8]

        # Create future for the result
        future: asyncio.Future = asyncio.get_event_loop().create_future()

        # Add to FIFO queue
        try:
            queue_size_before = self._submission_queue.qsize()
            self._submission_queue.put_nowait((coro, future, submission_id))
            self.logger().debug(
                f"[PIPELINE] Queued submission {submission_id} "
                f"(queue_size: {queue_size_before} -> {queue_size_before + 1})"
            )
        except asyncio.QueueFull:
            self.logger().error(
                f"[PIPELINE] Queue full! Rejecting submission {submission_id} "
                f"(max={self._max_queue_size})"
            )
            raise XRPLSystemBusyError("Pipeline queue is full, try again later")

        # Wait for result
        return await future

    async def _pipeline_loop(self):
        """
        Pipeline loop that serializes transaction submissions.

        Processes submissions one at a time with a configurable delay.
        Since only one transaction is processed at a time, autofill
        will always get the correct sequence number.
        """
        self.logger().debug("[PIPELINE] Loop started")

        while self._running:
            try:
                # Get next submission with timeout
                try:
                    coro, future, submission_id = await asyncio.wait_for(
                        self._submission_queue.get(),
                        timeout=1.0
                    )
                except asyncio.TimeoutError:
                    continue

                # Skip if future was already cancelled
                if future.done():
                    self.logger().debug(f"[PIPELINE] Skipping cancelled submission {submission_id}")
                    continue

                self._submissions_processed += 1
                queue_size = self._submission_queue.qsize()
                self.logger().debug(
                    f"[PIPELINE] Processing submission {submission_id} "
                    f"(#{self._submissions_processed}, queue_remaining={queue_size})"
                )

                # Execute the submission coroutine
                start_time = time.time()
                try:
                    result = await coro
                    elapsed_ms = (time.time() - start_time) * 1000
                    self._total_latency_ms += elapsed_ms

                    if not future.done():
                        future.set_result(result)

                    self.logger().debug(
                        f"[PIPELINE] Submission {submission_id} completed in {elapsed_ms:.1f}ms"
                    )

                except Exception as e:
                    elapsed_ms = (time.time() - start_time) * 1000
                    self._total_latency_ms += elapsed_ms
                    self._submissions_failed += 1

                    if not future.done():
                        future.set_exception(e)

                    self.logger().error(
                        f"[PIPELINE] Submission {submission_id} failed after {elapsed_ms:.1f}ms: {e}"
                    )

                # Delay before allowing next submission
                self.logger().debug(
                    f"[PIPELINE] Waiting {self._delay_seconds * 1000:.0f}ms before next submission"
                )
                await asyncio.sleep(self._delay_seconds)

            except asyncio.CancelledError:
                break
            except Exception as e:
                self.logger().error(f"[PIPELINE] Unexpected error: {e}")

        self.logger().debug(
            f"[PIPELINE] Loop stopped (processed {self._submissions_processed} submissions)"
        )



import os
from dataclasses import dataclass
from decimal import Decimal
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import aioprocessing

from hummingbot import root_path

if TYPE_CHECKING:
    from hummingbot import ClientConfigAdapter

_default_paths: Optional["GatewayPaths"] = None
_hummingbot_pipe: Optional[aioprocessing.AioConnection] = None

S_DECIMAL_0: Decimal = Decimal(0)


@dataclass
class GatewayPaths:
    """
    Represents the local paths and Docker mount paths for a gateway container's conf, certs and logs directories.

    Local paths represent where Hummingbot client sees the paths from the perspective of its local environment. If
    Hummingbot is being run from source, then the local environment is the same as the host environment. However, if
    Hummingbot is being run as a container, then the local environment is the container's environment.

    Mount paths represent where the gateway container's paths are located on the host environment. If Hummingbot is
    being run from source, then these should be the same as the local paths. However, if Hummingbot is being run as a
    container - then these must be fed to it from external sources (e.g. environment variables), since containers
    generally only have very restricted access to the host filesystem.
    """

    local_conf_path: Path
    local_certs_path: Path
    local_logs_path: Path
    mount_conf_path: Path
    mount_certs_path: Path
    mount_logs_path: Path

    def __post_init__(self):
        """
        Ensure the local paths are created when a GatewayPaths object is created.
        """
        for path in [self.local_conf_path, self.local_certs_path, self.local_logs_path]:
            path.mkdir(mode=0o755, parents=True, exist_ok=True)


def get_gateway_paths(client_config_map: "ClientConfigAdapter") -> GatewayPaths:
    """
    Calculates the default paths for a gateway container.

    For Hummingbot running from source, the gateway files are to be stored in ~/.hummingbot-gateway/<container name>/

    For Hummingbot running inside container, the gateway files are to be stored in ~/.hummingbot-gateway/ locally;
      and inside the paths pointed to be CERTS_FOLDER, GATEWAY_CONF_FOLDER, GATEWAY_LOGS_FOLDER environment variables
      on the host system.
    """
    global _default_paths
    if _default_paths is not None:
        return _default_paths

    external_certs_path: Optional[Path] = os.getenv("CERTS_FOLDER") and Path(os.getenv("CERTS_FOLDER"))
    external_conf_path: Optional[Path] = os.getenv("GATEWAY_CONF_FOLDER") and Path(os.getenv("GATEWAY_CONF_FOLDER"))
    external_logs_path: Optional[Path] = os.getenv("GATEWAY_LOGS_FOLDER") and Path(os.getenv("GATEWAY_LOGS_FOLDER"))
    local_certs_path: Path = root_path().joinpath("certs")
    local_conf_path: Path = root_path().joinpath("gateway/conf")
    local_logs_path: Path = root_path().joinpath("gateway/logs")
    mount_certs_path: Path = external_certs_path or local_certs_path
    mount_conf_path: Path = external_conf_path or local_conf_path
    mount_logs_path: Path = external_logs_path or local_logs_path

    _default_paths = GatewayPaths(
        local_conf_path=local_conf_path,
        local_certs_path=local_certs_path,
        local_logs_path=local_logs_path,
        mount_conf_path=mount_conf_path,
        mount_certs_path=mount_certs_path,
        mount_logs_path=mount_logs_path
    )
    return _default_paths



# distutils: language=c++

cdef extern from "<iterator>" namespace "std" nogil:
    cdef cppclass reverse_iterator[T]:
        pass

cdef extern from "cpp/Utils.h":
    cdef const T getIteratorFromReverseIterator[T](const reverse_iterator[T] rit)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Side(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    BID: _ClassVar[Side]
    ASK: _ClassVar[Side]

class KlineInterval(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    S1: _ClassVar[KlineInterval]
    M1: _ClassVar[KlineInterval]
    M15: _ClassVar[KlineInterval]
    H1: _ClassVar[KlineInterval]
    H4: _ClassVar[KlineInterval]
    D1: _ClassVar[KlineInterval]

class RateUpdateSide(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    BASE: _ClassVar[RateUpdateSide]
    QUOTE: _ClassVar[RateUpdateSide]
BID: Side
ASK: Side
S1: KlineInterval
M1: KlineInterval
M15: KlineInterval
H1: KlineInterval
H4: KlineInterval
D1: KlineInterval
BASE: RateUpdateSide
QUOTE: RateUpdateSide

class MdMessage(_message.Message):
    __slots__ = ("heartbeat", "summary", "trades", "mbo_snapshot", "mbo_diff", "mbp_snapshot", "mbp_diff", "kline")
    HEARTBEAT_FIELD_NUMBER: _ClassVar[int]
    SUMMARY_FIELD_NUMBER: _ClassVar[int]
    TRADES_FIELD_NUMBER: _ClassVar[int]
    MBO_SNAPSHOT_FIELD_NUMBER: _ClassVar[int]
    MBO_DIFF_FIELD_NUMBER: _ClassVar[int]
    MBP_SNAPSHOT_FIELD_NUMBER: _ClassVar[int]
    MBP_DIFF_FIELD_NUMBER: _ClassVar[int]
    KLINE_FIELD_NUMBER: _ClassVar[int]
    heartbeat: Heartbeat
    summary: Summary
    trades: Trades
    mbo_snapshot: MarketByOrder
    mbo_diff: MarketByOrderDiff
    mbp_snapshot: MarketByPrice
    mbp_diff: MarketByPriceDiff
    kline: Kline
    def __init__(self, heartbeat: _Optional[_Union[Heartbeat, _Mapping]] = ..., summary: _Optional[_Union[Summary, _Mapping]] = ..., trades: _Optional[_Union[Trades, _Mapping]] = ..., mbo_snapshot: _Optional[_Union[MarketByOrder, _Mapping]] = ..., mbo_diff: _Optional[_Union[MarketByOrderDiff, _Mapping]] = ..., mbp_snapshot: _Optional[_Union[MarketByPrice, _Mapping]] = ..., mbp_diff: _Optional[_Union[MarketByPriceDiff, _Mapping]] = ..., kline: _Optional[_Union[Kline, _Mapping]] = ...) -> None: ...

class MarketByPrice(_message.Message):
    __slots__ = ("levels", "chunk", "num_chunks")
    class Level(_message.Message):
        __slots__ = ("price", "quantity", "side")
        PRICE_FIELD_NUMBER: _ClassVar[int]
        QUANTITY_FIELD_NUMBER: _ClassVar[int]
        SIDE_FIELD_NUMBER: _ClassVar[int]
        price: int
        quantity: int
        side: Side
        def __init__(self, price: _Optional[int] = ..., quantity: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ...) -> None: ...
    LEVELS_FIELD_NUMBER: _ClassVar[int]
    CHUNK_FIELD_NUMBER: _ClassVar[int]
    NUM_CHUNKS_FIELD_NUMBER: _ClassVar[int]
    levels: _containers.RepeatedCompositeFieldContainer[MarketByPrice.Level]
    chunk: int
    num_chunks: int
    def __init__(self, levels: _Optional[_Iterable[_Union[MarketByPrice.Level, _Mapping]]] = ..., chunk: _Optional[int] = ..., num_chunks: _Optional[int] = ...) -> None: ...

class MarketByPriceDiff(_message.Message):
    __slots__ = ("diffs", "total_bid_levels", "total_ask_levels")
    class DiffOp(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ADD: _ClassVar[MarketByPriceDiff.DiffOp]
        REMOVE: _ClassVar[MarketByPriceDiff.DiffOp]
        REPLACE: _ClassVar[MarketByPriceDiff.DiffOp]
    ADD: MarketByPriceDiff.DiffOp
    REMOVE: MarketByPriceDiff.DiffOp
    REPLACE: MarketByPriceDiff.DiffOp
    class Diff(_message.Message):
        __slots__ = ("price", "quantity", "side", "op")
        PRICE_FIELD_NUMBER: _ClassVar[int]
        QUANTITY_FIELD_NUMBER: _ClassVar[int]
        SIDE_FIELD_NUMBER: _ClassVar[int]
        OP_FIELD_NUMBER: _ClassVar[int]
        price: int
        quantity: int
        side: Side
        op: MarketByPriceDiff.DiffOp
        def __init__(self, price: _Optional[int] = ..., quantity: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ..., op: _Optional[_Union[MarketByPriceDiff.DiffOp, str]] = ...) -> None: ...
    DIFFS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_BID_LEVELS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_ASK_LEVELS_FIELD_NUMBER: _ClassVar[int]
    diffs: _containers.RepeatedCompositeFieldContainer[MarketByPriceDiff.Diff]
    total_bid_levels: int
    total_ask_levels: int
    def __init__(self, diffs: _Optional[_Iterable[_Union[MarketByPriceDiff.Diff, _Mapping]]] = ..., total_bid_levels: _Optional[int] = ..., total_ask_levels: _Optional[int] = ...) -> None: ...

class MarketByOrder(_message.Message):
    __slots__ = ("orders", "chunk", "num_chunks")
    class Order(_message.Message):
        __slots__ = ("price", "quantity", "exchange_order_id", "side", "priority")
        PRICE_FIELD_NUMBER: _ClassVar[int]
        QUANTITY_FIELD_NUMBER: _ClassVar[int]
        EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
        SIDE_FIELD_NUMBER: _ClassVar[int]
        PRIORITY_FIELD_NUMBER: _ClassVar[int]
        price: int
        quantity: int
        exchange_order_id: int
        side: Side
        priority: int
        def __init__(self, price: _Optional[int] = ..., quantity: _Optional[int] = ..., exchange_order_id: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ..., priority: _Optional[int] = ...) -> None: ...
    ORDERS_FIELD_NUMBER: _ClassVar[int]
    CHUNK_FIELD_NUMBER: _ClassVar[int]
    NUM_CHUNKS_FIELD_NUMBER: _ClassVar[int]
    orders: _containers.RepeatedCompositeFieldContainer[MarketByOrder.Order]
    chunk: int
    num_chunks: int
    def __init__(self, orders: _Optional[_Iterable[_Union[MarketByOrder.Order, _Mapping]]] = ..., chunk: _Optional[int] = ..., num_chunks: _Optional[int] = ...) -> None: ...

class MarketByOrderDiff(_message.Message):
    __slots__ = ("diffs", "total_bid_levels", "total_ask_levels", "total_bid_orders", "total_ask_orders")
    class DiffOp(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ADD: _ClassVar[MarketByOrderDiff.DiffOp]
        REMOVE: _ClassVar[MarketByOrderDiff.DiffOp]
        REPLACE: _ClassVar[MarketByOrderDiff.DiffOp]
    ADD: MarketByOrderDiff.DiffOp
    REMOVE: MarketByOrderDiff.DiffOp
    REPLACE: MarketByOrderDiff.DiffOp
    class Diff(_message.Message):
        __slots__ = ("price", "quantity", "exchange_order_id", "side", "op", "priority")
        PRICE_FIELD_NUMBER: _ClassVar[int]
        QUANTITY_FIELD_NUMBER: _ClassVar[int]
        EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
        SIDE_FIELD_NUMBER: _ClassVar[int]
        OP_FIELD_NUMBER: _ClassVar[int]
        PRIORITY_FIELD_NUMBER: _ClassVar[int]
        price: int
        quantity: int
        exchange_order_id: int
        side: Side
        op: MarketByOrderDiff.DiffOp
        priority: int
        def __init__(self, price: _Optional[int] = ..., quantity: _Optional[int] = ..., exchange_order_id: _Optional[int] = ..., side: _Optional[_Union[Side, str]] = ..., op: _Optional[_Union[MarketByOrderDiff.DiffOp, str]] = ..., priority: _Optional[int] = ...) -> None: ...
    DIFFS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_BID_LEVELS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_ASK_LEVELS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_BID_ORDERS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_ASK_ORDERS_FIELD_NUMBER: _ClassVar[int]
    diffs: _containers.RepeatedCompositeFieldContainer[MarketByOrderDiff.Diff]
    total_bid_levels: int
    total_ask_levels: int
    total_bid_orders: int
    total_ask_orders: int
    def __init__(self, diffs: _Optional[_Iterable[_Union[MarketByOrderDiff.Diff, _Mapping]]] = ..., total_bid_levels: _Optional[int] = ..., total_ask_levels: _Optional[int] = ..., total_bid_orders: _Optional[int] = ..., total_ask_orders: _Optional[int] = ...) -> None: ...

class Trades(_message.Message):
    __slots__ = ("trades",)
    class Trade(_message.Message):
        __slots__ = ("tradeId", "price", "aggressing_side", "resting_exchange_order_id", "fill_quantity", "transact_time", "aggressing_exchange_order_id")
        TRADEID_FIELD_NUMBER: _ClassVar[int]
        PRICE_FIELD_NUMBER: _ClassVar[int]
        AGGRESSING_SIDE_FIELD_NUMBER: _ClassVar[int]
        RESTING_EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
        FILL_QUANTITY_FIELD_NUMBER: _ClassVar[int]
        TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
        AGGRESSING_EXCHANGE_ORDER_ID_FIELD_NUMBER: _ClassVar[int]
        tradeId: int
        price: int
        aggressing_side: Side
        resting_exchange_order_id: int
        fill_quantity: int
        transact_time: int
        aggressing_exchange_order_id: int
        def __init__(self, tradeId: _Optional[int] = ..., price: _Optional[int] = ..., aggressing_side: _Optional[_Union[Side, str]] = ..., resting_exchange_order_id: _Optional[int] = ..., fill_quantity: _Optional[int] = ..., transact_time: _Optional[int] = ..., aggressing_exchange_order_id: _Optional[int] = ...) -> None: ...
    TRADES_FIELD_NUMBER: _ClassVar[int]
    trades: _containers.RepeatedCompositeFieldContainer[Trades.Trade]
    def __init__(self, trades: _Optional[_Iterable[_Union[Trades.Trade, _Mapping]]] = ...) -> None: ...

class Summary(_message.Message):
    __slots__ = ("open", "close", "low", "high", "base_volume_lo", "base_volume_hi", "quote_volume_lo", "quote_volume_hi")
    OPEN_FIELD_NUMBER: _ClassVar[int]
    CLOSE_FIELD_NUMBER: _ClassVar[int]
    LOW_FIELD_NUMBER: _ClassVar[int]
    HIGH_FIELD_NUMBER: _ClassVar[int]
    BASE_VOLUME_LO_FIELD_NUMBER: _ClassVar[int]
    BASE_VOLUME_HI_FIELD_NUMBER: _ClassVar[int]
    QUOTE_VOLUME_LO_FIELD_NUMBER: _ClassVar[int]
    QUOTE_VOLUME_HI_FIELD_NUMBER: _ClassVar[int]
    open: int
    close: int
    low: int
    high: int
    base_volume_lo: int
    base_volume_hi: int
    quote_volume_lo: int
    quote_volume_hi: int
    def __init__(self, open: _Optional[int] = ..., close: _Optional[int] = ..., low: _Optional[int] = ..., high: _Optional[int] = ..., base_volume_lo: _Optional[int] = ..., base_volume_hi: _Optional[int] = ..., quote_volume_lo: _Optional[int] = ..., quote_volume_hi: _Optional[int] = ...) -> None: ...

class Kline(_message.Message):
    __slots__ = ("interval", "start_time", "open", "close", "high", "low", "volume_lo", "volume_hi")
    INTERVAL_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    OPEN_FIELD_NUMBER: _ClassVar[int]
    CLOSE_FIELD_NUMBER: _ClassVar[int]
    HIGH_FIELD_NUMBER: _ClassVar[int]
    LOW_FIELD_NUMBER: _ClassVar[int]
    VOLUME_LO_FIELD_NUMBER: _ClassVar[int]
    VOLUME_HI_FIELD_NUMBER: _ClassVar[int]
    interval: KlineInterval
    start_time: int
    open: int
    close: int
    high: int
    low: int
    volume_lo: int
    volume_hi: int
    def __init__(self, interval: _Optional[_Union[KlineInterval, str]] = ..., start_time: _Optional[int] = ..., open: _Optional[int] = ..., close: _Optional[int] = ..., high: _Optional[int] = ..., low: _Optional[int] = ..., volume_lo: _Optional[int] = ..., volume_hi: _Optional[int] = ...) -> None: ...

class Heartbeat(_message.Message):
    __slots__ = ("request_id", "timestamp")
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    request_id: int
    timestamp: int
    def __init__(self, request_id: _Optional[int] = ..., timestamp: _Optional[int] = ...) -> None: ...

class MdMessages(_message.Message):
    __slots__ = ("messages",)
    MESSAGES_FIELD_NUMBER: _ClassVar[int]
    messages: _containers.RepeatedCompositeFieldContainer[MdMessage]
    def __init__(self, messages: _Optional[_Iterable[_Union[MdMessage, _Mapping]]] = ...) -> None: ...

class AggMessage(_message.Message):
    __slots__ = ("heartbeat", "top_of_books", "rate_updates")
    HEARTBEAT_FIELD_NUMBER: _ClassVar[int]
    TOP_OF_BOOKS_FIELD_NUMBER: _ClassVar[int]
    RATE_UPDATES_FIELD_NUMBER: _ClassVar[int]
    heartbeat: Heartbeat
    top_of_books: TopOfBooks
    rate_updates: RateUpdates
    def __init__(self, heartbeat: _Optional[_Union[Heartbeat, _Mapping]] = ..., top_of_books: _Optional[_Union[TopOfBooks, _Mapping]] = ..., rate_updates: _Optional[_Union[RateUpdates, _Mapping]] = ...) -> None: ...

class TopOfBook(_message.Message):
    __slots__ = ("market_id", "transact_time", "bid_price", "bid_quantity", "ask_price", "ask_quantity", "last_price", "rolling24h_price")
    MARKET_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSACT_TIME_FIELD_NUMBER: _ClassVar[int]
    BID_PRICE_FIELD_NUMBER: _ClassVar[int]
    BID_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    ASK_PRICE_FIELD_NUMBER: _ClassVar[int]
    ASK_QUANTITY_FIELD_NUMBER: _ClassVar[int]
    LAST_PRICE_FIELD_NUMBER: _ClassVar[int]
    ROLLING24H_PRICE_FIELD_NUMBER: _ClassVar[int]
    market_id: int
    transact_time: int
    bid_price: int
    bid_quantity: int
    ask_price: int
    ask_quantity: int
    last_price: int
    rolling24h_price: int
    def __init__(self, market_id: _Optional[int] = ..., transact_time: _Optional[int] = ..., bid_price: _Optional[int] = ..., bid_quantity: _Optional[int] = ..., ask_price: _Optional[int] = ..., ask_quantity: _Optional[int] = ..., last_price: _Optional[int] = ..., rolling24h_price: _Optional[int] = ...) -> None: ...

class TopOfBooks(_message.Message):
    __slots__ = ("tops",)
    TOPS_FIELD_NUMBER: _ClassVar[int]
    tops: _containers.RepeatedCompositeFieldContainer[TopOfBook]
    def __init__(self, tops: _Optional[_Iterable[_Union[TopOfBook, _Mapping]]] = ...) -> None: ...

class RateUpdate(_message.Message):
    __slots__ = ("asset_id", "timestamp", "rate", "side")
    ASSET_ID_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    RATE_FIELD_NUMBER: _ClassVar[int]
    SIDE_FIELD_NUMBER: _ClassVar[int]
    asset_id: int
    timestamp: int
    rate: int
    side: RateUpdateSide
    def __init__(self, asset_id: _Optional[int] = ..., timestamp: _Optional[int] = ..., rate: _Optional[int] = ..., side: _Optional[_Union[RateUpdateSide, str]] = ...) -> None: ...

class RateUpdates(_message.Message):
    __slots__ = ("updates",)
    UPDATES_FIELD_NUMBER: _ClassVar[int]
    updates: _containers.RepeatedCompositeFieldContainer[RateUpdate]
    def __init__(self, updates: _Optional[_Iterable[_Union[RateUpdate, _Mapping]]] = ...) -> None: ...

class ClientMessage(_message.Message):
    __slots__ = ("heartbeat", "config")
    HEARTBEAT_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    heartbeat: Heartbeat
    config: Config
    def __init__(self, heartbeat: _Optional[_Union[Heartbeat, _Mapping]] = ..., config: _Optional[_Union[Config, _Mapping]] = ...) -> None: ...

class Config(_message.Message):
    __slots__ = ("mbp", "mbo", "trades", "summary", "klines")
    MBP_FIELD_NUMBER: _ClassVar[int]
    MBO_FIELD_NUMBER: _ClassVar[int]
    TRADES_FIELD_NUMBER: _ClassVar[int]
    SUMMARY_FIELD_NUMBER: _ClassVar[int]
    KLINES_FIELD_NUMBER: _ClassVar[int]
    mbp: bool
    mbo: bool
    trades: bool
    summary: bool
    klines: _containers.RepeatedScalarFieldContainer[KlineInterval]
    def __init__(self, mbp: bool = ..., mbo: bool = ..., trades: bool = ..., summary: bool = ..., klines: _Optional[_Iterable[_Union[KlineInterval, str]]] = ...) -> None: ...



import asyncio
import logging
from decimal import Decimal
from typing import Optional

import aiohttp

from hummingbot.core.network_base import NetworkBase
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger


class CustomAPIDataFeed(NetworkBase):
    cadf_logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls.cadf_logger is None:
            cls.cadf_logger = logging.getLogger(__name__)
        return cls.cadf_logger

    def __init__(self, api_url, update_interval: float = 5.0):
        super().__init__()
        self._ready_event = asyncio.Event()
        self._shared_client: Optional[aiohttp.ClientSession] = None
        self._api_url = api_url
        self._check_network_interval = 30.0
        self._ev_loop = asyncio.get_event_loop()
        self._price: Decimal = Decimal("0")
        self._update_interval: float = update_interval
        self._fetch_price_task: Optional[asyncio.Task] = None

    @property
    def name(self):
        return "custom_api"

    @property
    def health_check_endpoint(self):
        return self._api_url

    def _http_client(self) -> aiohttp.ClientSession:
        if self._shared_client is None:
            self._shared_client = aiohttp.ClientSession()
        return self._shared_client

    async def check_network(self) -> NetworkStatus:
        client = self._http_client()
        async with client.request("GET", self.health_check_endpoint) as resp:
            status_text = await resp.text()
            if resp.status != 200:
                raise Exception(f"Custom API Feed {self.name} server error: {status_text}")
        return NetworkStatus.CONNECTED

    def get_price(self) -> Decimal:
        return self._price

    async def fetch_price_loop(self):
        while True:
            try:
                await self.fetch_price()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(f"Error fetching a new price from {self._api_url}.", exc_info=True,
                                      app_warning_msg="Couldn't fetch newest price from CustomAPI. "
                                                      "Check network connection.")

            await asyncio.sleep(self._update_interval)

    async def fetch_price(self):
        client = self._http_client()
        async with client.request("GET", self._api_url) as resp:
            resp_text = await resp.text()
            if resp.status != 200:
                raise Exception(f"Custom API Feed {self.name} server error: {resp_text}")
            self._price = Decimal(str(resp_text))
        self._ready_event.set()

    async def start_network(self):
        await self.stop_network()
        self._fetch_price_task = safe_ensure_future(self.fetch_price_loop())

    async def stop_network(self):
        if self._fetch_price_task is not None:
            self._fetch_price_task.cancel()
            self._fetch_price_task = None

    def start(self):
        NetworkBase.start(self)

    def stop(self):
        NetworkBase.stop(self)



import asyncio
import importlib
import inspect
import logging
import sys
import time
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union

from sqlalchemy.orm import Query, Session

from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_data_types import BaseClientModel
from hummingbot.client.config.config_helpers import ClientConfigAdapter, get_strategy_starter_file
from hummingbot.client.config.strategy_config_data_types import BaseStrategyConfigMap
from hummingbot.client.performance import PerformanceMetrics
from hummingbot.client.settings import SCRIPT_STRATEGIES_MODULE, STRATEGIES
from hummingbot.connector.connector_metrics_collector import DummyMetricsCollector, MetricsCollector
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.connector.markets_recorder import MarketsRecorder
from hummingbot.core.clock import Clock, ClockMode
from hummingbot.core.connector_manager import ConnectorManager
from hummingbot.core.gateway.gateway_http_client import GatewayHttpClient
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.kill_switch import KillSwitch
from hummingbot.exceptions import InvalidScriptModule
from hummingbot.logger import HummingbotLogger
from hummingbot.model.sql_connection_manager import SQLConnectionManager
from hummingbot.model.trade_fill import TradeFill
from hummingbot.notifier.notifier_base import NotifierBase
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.strategy_base import StrategyBase
from hummingbot.strategy.strategy_v2_base import StrategyV2Base, StrategyV2ConfigBase

# Constants
s_decimal_0 = Decimal("0")


class StrategyType(Enum):
    V2 = "v2"
    REGULAR = "regular"


s_logger = None


class TradingCore:
    """
    Core trading functionality with modular architecture.

    This class provides:
    - Connector management (create, add, remove connectors)
    - Market data access (order books, balances, etc.)
    - Strategy management (optional - can run without strategies)
    - Direct trading capabilities
    - Clock management for real-time operations
    """

    KILL_TIMEOUT = 20.0

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global s_logger
        if s_logger is None:
            s_logger = logging.getLogger(__name__)
        return s_logger

    def __init__(self,
                 client_config: Union[ClientConfigMap, ClientConfigAdapter, Dict[str, Any]],
                 scripts_path: Optional[Path] = None):
        """
        Initialize the trading core.

        Args:
            client_config: Configuration object or dictionary
            scripts_path: Optional path to script strategies directory
        """
        # Convert config to ClientConfigAdapter if needed
        if isinstance(client_config, dict):
            self.client_config_map = self._create_config_adapter_from_dict(client_config)
        elif isinstance(client_config, ClientConfigMap):
            self.client_config_map = ClientConfigAdapter(client_config)
        else:
            self.client_config_map = client_config

        # Strategy paths
        self.scripts_path = scripts_path or Path("scripts")

        # Core components
        self.connector_manager = ConnectorManager(self.client_config_map)
        self.clock: Optional[Clock] = None

        # Strategy components (optional)
        self.strategy: Optional[StrategyBase] = None
        self.strategy_name: Optional[str] = None
        self.strategy_config_map: Optional[BaseStrategyConfigMap] = None
        self.strategy_task: Optional[asyncio.Task] = None
        self._strategy_file_name: Optional[str] = None

        # Supporting components
        self.notifiers: List[NotifierBase] = []
        self.kill_switch: Optional[KillSwitch] = None
        self.markets_recorder: Optional[MarketsRecorder] = None
        self.trade_fill_db: Optional[SQLConnectionManager] = None

        # Metrics collectors mapping (connector_name -> MetricsCollector)
        self._metrics_collectors: Dict[str, MetricsCollector] = {}

        # Runtime state
        self.init_time: float = time.time()
        self.start_time: Optional[float] = None
        self._is_running: bool = False
        self._strategy_running: bool = False
        self._trading_required: bool = True

        # Config storage for flexible config loading
        self._config_source: Optional[str] = None
        self._config_data: Optional[Dict[str, Any]] = None

        # Backward compatibility properties
        self.market_trading_pairs_map: Dict[str, List[str]] = {}
        self.market_trading_pair_tuples: List[MarketTradingPairTuple] = []
        self._gateway_monitor = GatewayHttpClient.get_instance(self.client_config_map.hb_config.gateway)
        self._gateway_monitor.start_monitor()

    def _create_config_adapter_from_dict(self, config_dict: Dict[str, Any]) -> ClientConfigAdapter:
        """Create a ClientConfigAdapter from a dictionary."""
        client_config = ClientConfigMap()

        # Set configuration values
        for key, value in config_dict.items():
            if hasattr(client_config, key):
                setattr(client_config, key, value)

        return ClientConfigAdapter(client_config)

    @property
    def gateway_monitor(self):
        """Get the gateway monitor instance."""
        return self._gateway_monitor

    @property
    def markets(self) -> Dict[str, ExchangeBase]:
        """Get all markets/connectors (backward compatibility)."""
        return self.connector_manager.get_all_connectors()

    @property
    def connectors(self) -> Dict[str, ExchangeBase]:
        """Get all connectors (backward compatibility)."""
        return self.connector_manager.connectors

    @property
    def strategy_file_name(self) -> Optional[str]:
        """Get the strategy file name."""
        return self._strategy_file_name

    @strategy_file_name.setter
    def strategy_file_name(self, value: Optional[str]):
        """Set the strategy file name."""
        self._strategy_file_name = value

    async def start_clock(self) -> bool:
        """
        Start the clock system without requiring a strategy.

        This allows real-time market data updates and order management
        without needing an active strategy.
        """
        if self.clock is not None:
            self.logger().warning("Clock is already running")
            return False

        try:
            tick_size = self.client_config_map.tick_size
            self.logger().info(f"Creating the clock with tick size: {tick_size}")
            self.clock = Clock(ClockMode.REALTIME, tick_size=tick_size)

            # Add all connectors to clock
            for connector in self.connector_manager.connectors.values():
                if connector is not None:
                    self.clock.add_iterator(connector)
                    # Cancel dangling orders
                    if len(connector.limit_orders) > 0:
                        self.logger().info(f"Canceling dangling limit orders on {connector.name}...")
                        await connector.cancel_all(10.0)

            # Start the clock
            self._clock_task = asyncio.create_task(self._run_clock())
            self._is_running = True
            self.start_time = time.time() * 1e3

            self.logger().info("Clock started successfully")
            return True

        except Exception as e:
            self.logger().error(f"Failed to start clock: {e}")
            return False

    async def stop_clock(self) -> bool:
        """Stop the clock system."""
        if self.clock is None:
            return True

        try:
            # Cancel clock task
            if self._clock_task and not self._clock_task.done():
                self._clock_task.cancel()
                try:
                    await self._clock_task
                except asyncio.CancelledError:
                    pass

            self.clock = None
            self._is_running = False

            self.logger().info("Clock stopped successfully")
            return True

        except Exception as e:
            self.logger().error(f"Failed to stop clock: {e}")
            return False

    async def create_connector(self,
                               connector_name: str,
                               trading_pairs: List[str],
                               trading_required: bool = True,
                               api_keys: Optional[Dict[str, str]] = None) -> ExchangeBase:
        """
        Create a connector instance.

        Args:
            connector_name: Name of the connector
            trading_pairs: List of trading pairs
            trading_required: Whether trading is required
            api_keys: Optional API keys

        Returns:
            ExchangeBase: Created connector
        """
        connector = self.connector_manager.create_connector(
            connector_name, trading_pairs, trading_required, api_keys
        )

        # Add to clock if running
        if self.clock and connector:
            self.clock.add_iterator(connector)

        # Add to markets recorder if exists
        if self.markets_recorder and connector:
            self.markets_recorder.add_market(connector)

        return connector

    def _initialize_metrics_for_connector(self, connector: ExchangeBase, connector_name: str):
        """Initialize metrics collector for a specific connector."""
        try:
            # Get the metrics collector from config
            collector = self.client_config_map.anonymized_metrics_mode.get_collector(
                connector=connector,
                rate_provider=RateOracle.get_instance(),
                instance_id=self.client_config_map.instance_id,
            )

            self.clock.add_iterator(collector)

            # Store the collector
            self._metrics_collectors[connector_name] = collector

            self.logger().debug(f"Metrics collector initialized for {connector_name}")

        except Exception as e:
            self.logger().warning(f"Failed to initialize metrics collector for {connector_name}: {e}")
            # Use dummy collector as fallback
            self._metrics_collectors[connector_name] = DummyMetricsCollector()

    def remove_connector(self, connector_name: str) -> bool:
        """
        Remove a connector.

        Args:
            connector_name: Name of the connector to remove

        Returns:
            bool: True if successfully removed
        """
        connector = self.connector_manager.get_connector(connector_name)

        if connector:
            # Stop and remove metrics collector first
            if connector_name in self._metrics_collectors:
                collector = self._metrics_collectors[connector_name]

                # Remove from clock if it was added
                if self.clock:
                    self.clock.remove_iterator(collector)

                # Remove from mapping
                del self._metrics_collectors[connector_name]

                self.logger().debug(f"Metrics collector stopped for {connector_name}")

            # Remove from clock if exists
            connector.stop(self.clock)
            if self.clock:
                self.clock.remove_iterator(connector)

            # Remove from markets recorder if exists
            if self.markets_recorder:
                self.markets_recorder.remove_market(connector)

        return self.connector_manager.remove_connector(connector_name)

    def detect_strategy_type(self, strategy_name: str) -> StrategyType:
        """Detect the type of strategy."""
        if self.is_v2_strategy(strategy_name):
            return StrategyType.V2
        elif strategy_name in STRATEGIES:
            return StrategyType.REGULAR
        else:
            raise ValueError(f"Unknown strategy: {strategy_name}")

    def is_v2_strategy(self, strategy_name: str) -> bool:
        """Check if the strategy is a V2 strategy."""
        v2_file = self.scripts_path / f"{strategy_name}.py"
        return v2_file.exists()

    def initialize_markets_recorder(self, db_name: str = None):
        """
        Initialize markets recorder for trade persistence.

        Args:
            db_name: Database name (defaults to strategy file name)
        """
        if not db_name:
            # For V2 strategies with config files, use the config source as db name
            # Otherwise use strategy file name
            if self._config_source and self.is_v2_strategy(self.strategy_name or ""):
                db_name = self._config_source
            else:
                db_name = self._strategy_file_name or "trades"

        if db_name.endswith(".yml") or db_name.endswith(".py"):
            db_name = db_name.split(".")[0]

        self.trade_fill_db = SQLConnectionManager.get_trade_fills_instance(
            self.client_config_map, db_name
        )

        self.markets_recorder = MarketsRecorder(
            self.trade_fill_db,
            list(self.connector_manager.connectors.values()),
            self._strategy_file_name or db_name,
            self.strategy_name or db_name,
            self.client_config_map.market_data_collection
        )

        self.markets_recorder.start()
        self.logger().info(f"Markets recorder initialized with database: {db_name}")

    def load_v2_class(self, strategy_name: str) -> Tuple[Type, BaseClientModel]:
        """
        Load V2 strategy class and its config.

        Every V2 strategy must have a config class (subclass of StrategyV2ConfigBase).
        Config is always loaded - either from a YAML file or with defaults.

        Args:
            strategy_name: Name of the V2 strategy

        Returns:
            Tuple of (strategy_class, config_object)
        """
        module = sys.modules.get(f"{SCRIPT_STRATEGIES_MODULE}.{strategy_name}")

        if module is not None:
            strategy_module = importlib.reload(module)
        else:
            strategy_module = importlib.import_module(f".{strategy_name}", package=SCRIPT_STRATEGIES_MODULE)

        try:
            strategy_class = next((member for member_name, member in inspect.getmembers(strategy_module)
                                   if inspect.isclass(member) and
                                   issubclass(member, StrategyV2Base) and
                                   member is not StrategyV2Base))
        except StopIteration:
            raise InvalidScriptModule(f"The module {strategy_name} does not contain any subclass of StrategyV2Base")

        # Always load config class
        try:
            config_class = next((member for member_name, member in inspect.getmembers(strategy_module)
                                 if inspect.isclass(member) and
                                 issubclass(member, BaseClientModel) and
                                 member not in [BaseClientModel, StrategyV2ConfigBase]))
        except StopIteration:
            raise InvalidScriptModule(f"The module {strategy_name} does not contain any subclass of StrategyV2ConfigBase")

        # Load config data from file or use defaults
        config_data = self._load_strategy_config()
        config = config_class(**config_data)
        strategy_class.init_markets(config)

        return strategy_class, config

    def _load_strategy_config(self) -> Dict[str, Any]:
        """
        Load strategy configuration from various sources.

        This method can be overridden by subclasses to load from different sources
        (dict, database, remote API, etc.) instead of filesystem.
        """
        if self._config_data:
            return self._config_data
        elif self._config_source:
            # Load from YAML config file
            return self._load_v2_yaml_config(self._config_source)
        else:
            return {}

    def _load_v2_yaml_config(self, config_file_path: str) -> Dict[str, Any]:
        """Load YAML configuration file for V2 strategies."""
        import yaml

        from hummingbot.client.settings import SCRIPT_STRATEGY_CONF_DIR_PATH

        try:
            # Try direct path first
            if "/" in config_file_path or "\\" in config_file_path:
                config_path = Path(config_file_path)
            else:
                # Assume it's in the V2 strategy config directory
                config_path = SCRIPT_STRATEGY_CONF_DIR_PATH / config_file_path

            with open(config_path, 'r') as file:
                return yaml.safe_load(file)
        except Exception as e:
            self.logger().warning(f"Failed to load config file {config_file_path}: {e}")
            return {}

    async def start_strategy(self,
                             strategy_name: str,
                             strategy_config: Optional[Union[BaseStrategyConfigMap, Dict[str, Any], str]] = None,
                             strategy_file_name: Optional[str] = None) -> bool:
        """
        Start a trading strategy.

        Args:
            strategy_name: Name of the strategy
            strategy_config: Strategy configuration (object, dict, or file path)
            strategy_file_name: Optional file name for the strategy

        Returns:
            bool: True if strategy started successfully
        """
        try:
            if self._strategy_running:
                self.logger().warning("Strategy is already running")
                return False

            self.strategy_name = strategy_name
            self._strategy_file_name = strategy_file_name or strategy_name

            # Store config for later use
            if isinstance(strategy_config, str):
                # File path - will be loaded by _load_strategy_config
                self._config_source = strategy_config
            elif isinstance(strategy_config, dict):
                self._config_data = strategy_config

            # Initialize strategy based on type
            strategy_type = self.detect_strategy_type(strategy_name)

            if strategy_type == StrategyType.V2:
                await self._initialize_v2_strategy()
            else:
                await self._initialize_regular_strategy()

            # Initialize markets for backward compatibility
            self._initialize_markets_for_strategy()

            # Start the trading execution loop
            await self._start_strategy_execution()

            # Start rate oracle (required for PNL calculation)
            RateOracle.get_instance().start()

            self._strategy_running = True

            self.logger().info(f"Strategy {strategy_name} started successfully")
            return True

        except Exception as e:
            self.logger().error(f"Failed to start strategy {strategy_name}: {e}")
            return False

    async def _initialize_v2_strategy(self):
        """Initialize a V2 strategy using consolidated approach."""
        v2_strategy_class, config = self.load_v2_class(self.strategy_name)

        # Get markets from V2 class
        markets_list = [(conn, list(pairs)) for conn, pairs in v2_strategy_class.markets.items()]

        # Initialize markets using single method
        await self.initialize_markets(markets_list)

        # Create strategy instance (config is always present)
        self.strategy = v2_strategy_class(self.markets, config)

    async def _initialize_regular_strategy(self):
        """Initialize a regular strategy using starter file."""
        start_strategy_func: Callable = get_strategy_starter_file(self.strategy_name)
        if asyncio.iscoroutinefunction(start_strategy_func):
            await start_strategy_func(self)
        else:
            start_strategy_func(self)

    async def _start_strategy_execution(self):
        """
        Start the strategy execution system.
        """
        try:
            # Ensure markets recorder exists (should have been created during market initialization)
            if not self.markets_recorder:
                self.initialize_markets_recorder()

            # Ensure clock exists
            if self.clock is None:
                await self.start_clock()

            # Add strategy to clock
            if self.strategy and self.clock:
                self.clock.add_iterator(self.strategy)

                # Restore market states if markets recorder exists
                if self.markets_recorder:
                    for market in self.markets.values():
                        self.markets_recorder.restore_market_states(self._strategy_file_name, market)

                for connector_name, connector in self.connector_manager.connectors.items():
                    if connector_name not in self._metrics_collectors and "_paper_trade" not in connector_name:
                        self.logger().debug(f"Initializing metrics collector for {connector_name} (created outside normal flow)")
                        self._initialize_metrics_for_connector(connector, connector_name)

            # Initialize kill switch if enabled
            if (self._trading_required and
                    self.client_config_map.kill_switch_mode.model_config.get("title") == "kill_switch_enabled"):
                self.kill_switch = self.client_config_map.kill_switch_mode.get_kill_switch(self)
                await self._wait_till_ready(self.kill_switch.start)

            self.logger().info(f"'{self.strategy_name}' strategy execution started.")

        except Exception as e:
            self.logger().error(f"Error starting strategy execution: {e}", exc_info=True)
            raise

    async def _run_clock(self):
        """Run the clock system."""
        with self.clock as clock:
            await clock.run()

    async def _wait_till_ready(self, func: Callable, *args, **kwargs):
        """Wait until all markets are ready before executing function."""
        while True:
            all_ready = all([market.ready for market in self.markets.values()])
            if not all_ready:
                await asyncio.sleep(0.5)
            else:
                if inspect.iscoroutinefunction(func):
                    return await func(*args, **kwargs)
                return func(*args, **kwargs)

    async def stop_strategy(self) -> bool:
        """Stop the currently running strategy."""
        try:
            if not self._strategy_running:
                self.logger().warning("No strategy is currently running")
                return False

            # Remove strategy from clock FIRST to prevent further ticks
            if self.clock is not None and self.strategy is not None:
                self.clock.remove_iterator(self.strategy)

            # Remove kill switch from clock
            if self.clock is not None and self.kill_switch is not None:
                self.kill_switch.stop()

            # Stop rate oracle
            RateOracle.get_instance().stop()

            # Clean up strategy components
            self.strategy = None
            self.strategy_task = None
            self.kill_switch = None
            self._strategy_running = False

            self.logger().info("Strategy stopped successfully")
            return True

        except Exception as e:
            self.logger().error(f"Failed to stop strategy: {e}")
            return False

    async def cancel_outstanding_orders(self) -> bool:
        """Cancel all outstanding orders."""
        try:
            cancellation_tasks = []
            for connector in self.connector_manager.connectors.values():
                if len(connector.limit_orders) > 0:
                    cancellation_tasks.append(connector.cancel_all(self.KILL_TIMEOUT))

            if cancellation_tasks:
                await asyncio.gather(*cancellation_tasks, return_exceptions=True)

            return True
        except Exception as e:
            self.logger().error(f"Error cancelling orders: {e}")
            return False

    def _initialize_markets_for_strategy(self):
        """Initialize market data structures for backward compatibility."""
        # Update market trading pairs map
        self.market_trading_pairs_map.clear()
        for name, connector in self.connector_manager.connectors.items():
            self.market_trading_pairs_map[name] = connector.trading_pairs

        # Update market trading pair tuples
        self.market_trading_pair_tuples = [
            MarketTradingPairTuple(connector, trading_pair, base, quote)
            for name, connector in self.connector_manager.connectors.items()
            for trading_pair in connector.trading_pairs
            for base, quote in [trading_pair.split("-")]
        ]

    def get_status(self) -> Dict[str, Any]:
        """Get current status of the trading engine."""
        return {
            'clock_running': self._is_running,
            'strategy_running': self._strategy_running,
            'strategy_name': self.strategy_name,
            'strategy_file_name': self._strategy_file_name,
            'strategy_type': self.detect_strategy_type(self.strategy_name).value if self.strategy_name else None,
            'start_time': self.start_time,
            'uptime': (time.time() * 1e3 - self.start_time) if self.start_time else 0,
            'connectors': self.connector_manager.get_status(),
            'kill_switch_enabled': self.client_config_map.kill_switch_mode.model_config.get("title") == "kill_switch_enabled",
            'markets_recorder_active': self.markets_recorder is not None,
        }

    def add_notifier(self, notifier: NotifierBase):
        """Add a notifier to the engine."""
        self.notifiers.append(notifier)

    def notify(self, msg: str, level: str = "INFO"):
        """Send a notification."""
        self.logger().log(getattr(logging, level.upper(), logging.INFO), msg)
        for notifier in self.notifiers:
            notifier.add_message_to_queue(msg)

    async def initialize_markets(self, market_names: List[Tuple[str, List[str]]]):
        """
        Initialize markets - single method that works for all strategy types.

        This replaces all the redundant initialize_markets* methods with one consistent approach.

        Args:
            market_names: List of (exchange_name, trading_pairs) tuples
        """
        # Create connectors for each market
        for connector_name, trading_pairs in market_names:
            # for now we identify gateway connector that contain "/" in their name
            if "/" in connector_name:
                await self.gateway_monitor.wait_for_online_status()
            connector = self.connector_manager.create_connector(
                connector_name, trading_pairs, self._trading_required
            )

            # Add to clock if running
            if self.clock and connector:
                self.clock.add_iterator(connector)

        # Initialize markets recorder now that connectors exist
        if not self.markets_recorder:
            self.initialize_markets_recorder()

        # Add connectors to markets recorder
        if self.markets_recorder:
            for connector in self.connector_manager.connectors.values():
                self.markets_recorder.add_market(connector)

    def get_balance(self, connector_name: str, asset: str) -> float:
        """Get balance for an asset from a connector."""
        return self.connector_manager.get_balance(connector_name, asset)

    def get_order_book(self, connector_name: str, trading_pair: str):
        """Get order book from a connector."""
        return self.connector_manager.get_order_book(connector_name, trading_pair)

    async def get_current_balances(self, connector_name: str):
        if connector_name in self.connector_manager.connectors and self.connector_manager.connectors[connector_name].ready:
            return self.connector_manager.connectors[connector_name].get_all_balances()
        elif "Paper" in connector_name:
            paper_balances = self.client_config_map.paper_trade.paper_trade_account_balance
            if paper_balances is None:
                return {}
            return {token: Decimal(str(bal)) for token, bal in paper_balances.items()}
        else:
            await self.connector_manager.update_connector_balances(connector_name)
            return self.connector_manager.get_all_balances(connector_name)

    async def calculate_profitability(self) -> Decimal:
        """
        Determines the profitability of the trading bot.
        This function is used by the KillSwitch class.
        Must be updated if the method of performance report gets updated.
        """
        if not self.markets_recorder:
            return s_decimal_0
        if not self.trade_fill_db:
            return s_decimal_0
        if any(not market.ready for market in self.connector_manager.connectors.values()):
            return s_decimal_0

        start_time = self.init_time

        with self.trade_fill_db.get_new_session() as session:
            trades: List[TradeFill] = self._get_trades_from_session(
                int(start_time * 1e3),
                session=session,
                config_file_path=self.strategy_file_name)
            perf_metrics = await self.calculate_performance_metrics_by_connector_pair(trades)
            returns_pct = [perf.return_pct for perf in perf_metrics]
            return sum(returns_pct) / len(returns_pct) if len(returns_pct) > 0 else s_decimal_0

    async def calculate_performance_metrics_by_connector_pair(self, trades: List[TradeFill]) -> List[PerformanceMetrics]:
        """
        Calculates performance metrics by connector and trading pair using the provided trades and the PerformanceMetrics class.
        """
        market_info: Set[Tuple[str, str]] = set((t.market, t.symbol) for t in trades)
        performance_metrics: List[PerformanceMetrics] = []
        for market, symbol in market_info:
            cur_trades = [t for t in trades if t.market == market and t.symbol == symbol]
            network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
            try:
                cur_balances = await asyncio.wait_for(self.get_current_balances(market), network_timeout)
            except asyncio.TimeoutError:
                self.logger().warning("\nA network error prevented the balances retrieval to complete. See logs for more details.")
                raise
            perf = await PerformanceMetrics.create(symbol, cur_trades, cur_balances)
            performance_metrics.append(perf)
        return performance_metrics

    @staticmethod
    def _get_trades_from_session(start_timestamp: int,
                                 session: Session,
                                 number_of_rows: Optional[int] = None,
                                 config_file_path: str = None) -> List[TradeFill]:

        filters = [TradeFill.timestamp >= start_timestamp]
        if config_file_path is not None:
            filters.append(TradeFill.config_file_path.like(f"%{config_file_path}%"))
        query: Query = (session
                        .query(TradeFill)
                        .filter(*filters)
                        .order_by(TradeFill.timestamp.desc()))
        if number_of_rows is None:
            result: List[TradeFill] = query.all() or []
        else:
            result: List[TradeFill] = query.limit(number_of_rows).all() or []

        result.reverse()
        return result

    async def shutdown(self, skip_order_cancellation: bool = False) -> bool:
        """
        Shutdown the trading core completely.

        This stops all strategies, connectors, and the clock.

        Args:
            skip_order_cancellation: Whether to skip cancelling outstanding orders
        """
        try:
            # Handle V2 strategy specific cleanup first
            if self.strategy and isinstance(self.strategy, StrategyV2Base):
                await self.strategy.on_stop()

            # Stop strategy if running
            if self._strategy_running:
                await self.stop_strategy()

            # Cancel outstanding orders
            if not skip_order_cancellation:
                await self.cancel_outstanding_orders()

            # Stop all metrics collectors first
            for connector_name, collector in list(self._metrics_collectors.items()):
                try:
                    if self.clock:
                        self.clock.remove_iterator(collector)
                    self.logger().debug(f"Stopped metrics collector for {connector_name}")
                except Exception as e:
                    self.logger().error(f"Error stopping metrics collector for {connector_name}: {e}")

            self._metrics_collectors.clear()

            # Remove all connectors
            connector_names = list(self.connector_manager.connectors.keys())
            for name in connector_names:
                try:
                    self.remove_connector(name)
                except Exception as e:
                    self.logger().error(f"Error stopping connector {name}: {e}")

            # Stop clock if running
            if self._is_running:
                await self.stop_clock()

            # Stop markets recorder
            if self.markets_recorder:
                self.markets_recorder.stop()
                self.markets_recorder = None

            # Stop gateway monitor
            if self._gateway_monitor:
                self._gateway_monitor.stop_monitor()

            # Clear strategy references
            self.strategy = None
            self.strategy_name = None
            self.strategy_config_map = None
            self._strategy_file_name = None
            self._config_source = None
            self._config_data = None

            self.logger().info("Trading core shutdown complete")
            return True

        except Exception as e:
            self.logger().error(f"Error during shutdown: {e}")
            return False



cdef class InFlightOrderBase:
    cdef:
        public str client_order_id
        public str exchange_order_id
        public str trading_pair
        public object order_type
        public object trade_type
        public object price
        public object amount
        public object executed_amount_base
        public object executed_amount_quote
        public str fee_asset
        public object fee_paid
        public str last_state
        public object exchange_order_id_update_event
        public object completely_filled_event
        public double _creation_timestamp



import numpy as np

from .base_trailing_indicator import BaseTrailingIndicator


class HistoricalVolatilityIndicator(BaseTrailingIndicator):
    def __init__(self, sampling_length: int = 30, processing_length: int = 15):
        super().__init__(sampling_length, processing_length)

    def _indicator_calculation(self) -> float:
        prices = self._sampling_buffer.get_as_numpy_array()
        if prices.size > 0:
            log_returns = np.diff(np.log(prices))
            return np.var(log_returns)

    def _processing_calculation(self) -> float:
        processing_array = self._processing_buffer.get_as_numpy_array()
        if processing_array.size > 0:
            return np.sqrt(np.mean(np.nan_to_num(processing_array)))



import asyncio
import copy
from decimal import Decimal
from typing import Any, Dict, Optional, Tuple

from async_timeout import timeout

from hummingbot.core.data_type.common import OrderType, PositionAction, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate

GET_GATEWAY_EX_ORDER_ID_TIMEOUT = 30  # seconds
GET_GATEWAY_TX_HASH = 30  # seconds

s_decimal_0 = Decimal("0")


class GatewayInFlightOrder(InFlightOrder):
    def __init__(
        self,
        client_order_id: str,
        trading_pair: str,
        order_type: OrderType,
        trade_type: TradeType,
        creation_timestamp: float,
        price: Decimal = s_decimal_0,
        amount: Decimal = s_decimal_0,
        exchange_order_id: Optional[str] = None,
        creation_transaction_hash: Optional[str] = None,
        gas_price: Optional[Decimal] = s_decimal_0,
        initial_state: OrderState = OrderState.PENDING_CREATE,
        leverage: int = 1,
        position: PositionAction = PositionAction.NIL,
    ):
        super().__init__(
            client_order_id=client_order_id,
            exchange_order_id=exchange_order_id,
            trading_pair=trading_pair,
            order_type=order_type,
            trade_type=trade_type,
            price=price,
            amount=amount,
            creation_timestamp=creation_timestamp,
            initial_state=initial_state,
            leverage=leverage,
            position=position,
        )
        self._fee_asset = trading_pair.split("-")[0]  # defaults to base asset
        self._gas_price = gas_price
        self._nonce: int = -1
        self._creation_transaction_hash: Optional[str] = creation_transaction_hash
        self._creation_transaction_hash_update_event = asyncio.Event()
        if self.creation_transaction_hash is not None:
            self._creation_transaction_hash_update_event.set()
        self._cancel_tx_hash: Optional[str] = None

    @property
    def gas_price(self) -> Decimal:
        return self._gas_price

    @gas_price.setter
    def gas_price(self, gas_price: Decimal):
        self._gas_price = gas_price

    @property
    def fee_asset(self) -> str:
        return self._fee_asset

    @fee_asset.setter
    def fee_asset(self, fee_asset: str):
        self._fee_asset = fee_asset

    @property
    def nonce(self) -> int:
        return self._nonce

    @nonce.setter
    def nonce(self, nonce):
        self._nonce = nonce

    @property
    def creation_transaction_hash(self) -> Optional[str]:
        return self._creation_transaction_hash

    @creation_transaction_hash.setter
    def creation_transaction_hash(self, creation_transaction_hash: str):
        self._creation_transaction_hash = creation_transaction_hash

    @property
    def cancel_tx_hash(self) -> Optional[str]:
        return self._cancel_tx_hash

    @cancel_tx_hash.setter
    def cancel_tx_hash(self, cancel_tx_hash):
        self._cancel_tx_hash = cancel_tx_hash

    async def get_exchange_order_id(self) -> Optional[str]:
        """
        Overridden from parent class because blockchain orders take more time than ones from CEX.
        """
        if self.exchange_order_id is None:
            async with timeout(GET_GATEWAY_EX_ORDER_ID_TIMEOUT):
                await self.exchange_order_id_update_event.wait()
        return self.exchange_order_id

    @property
    def attributes(self) -> Tuple[Any]:
        return copy.deepcopy(
            (
                self.client_order_id,
                self.trading_pair,
                self.order_type,
                self.trade_type,
                self.price,
                self.amount,
                self.exchange_order_id,
                self.current_state,
                self.leverage,
                self.position,
                self.executed_amount_base,
                self.executed_amount_quote,
                self.creation_timestamp,
                self.last_update_timestamp,
                self.nonce,
                self.gas_price,
                self._creation_transaction_hash,
                self.cancel_tx_hash,
            )
        )

    @property
    def is_done(self) -> bool:
        return super().is_done

    def update_creation_transaction_hash(self, creation_transaction_hash: str):
        self.creation_transaction_hash = creation_transaction_hash
        self._creation_transaction_hash_update_event.set()

    async def get_creation_transaction_hash(self) -> str:
        if self.creation_transaction_hash is None:
            async with timeout(GET_GATEWAY_TX_HASH):
                await self._creation_transaction_hash_update_event.wait()
        return self.creation_transaction_hash

    def update_with_order_update(self, order_update: OrderUpdate) -> bool:
        """
        Updates the in flight order with an order update
        return: True if the order gets updated otherwise False
        """
        if (
            order_update.client_order_id != self.client_order_id
            and order_update.exchange_order_id != self.exchange_order_id
        ):
            return False

        prev_data = self.attributes

        if self.exchange_order_id is None and order_update.exchange_order_id is not None:
            self.update_exchange_order_id(order_update.exchange_order_id)

        self.current_state = order_update.new_state
        self.check_processed_by_exchange_condition()
        misc_updates = order_update.misc_updates or {}
        creation_transaction_hash = misc_updates.get("creation_transaction_hash", self.creation_transaction_hash)
        if creation_transaction_hash is not None:
            self.update_creation_transaction_hash(creation_transaction_hash=creation_transaction_hash)
        self._cancel_tx_hash = misc_updates.get("cancelation_transaction_hash", self._cancel_tx_hash)
        if self.current_state not in {OrderState.PENDING_CANCEL, OrderState.CANCELED}:
            if "fee_asset" in misc_updates:
                self.fee_asset = misc_updates["fee_asset"]

        updated: bool = prev_data != self.attributes

        if updated:
            self.last_update_timestamp = order_update.update_timestamp

        return updated

    @classmethod
    def from_json(cls, data: Dict[str, Any]) -> "GatewayInFlightOrder":
        """
        Initialize an InFlightOrder using a JSON object
        :param data: JSON data
        :return: Formatted InFlightOrder
        """
        order = GatewayInFlightOrder(
            client_order_id=data["client_order_id"],
            trading_pair=data["trading_pair"],
            order_type=getattr(OrderType, data["order_type"]),
            trade_type=getattr(TradeType, data["trade_type"]),
            amount=Decimal(data["amount"]),
            price=Decimal(data["price"]),
            exchange_order_id=data["exchange_order_id"],
            initial_state=OrderState(int(data["last_state"])),
            leverage=int(data["leverage"]),
            position=PositionAction(data["position"]),
            creation_timestamp=data.get("creation_timestamp", -1),
        )
        order.executed_amount_base = Decimal(data["executed_amount_base"])
        order.executed_amount_quote = Decimal(data["executed_amount_quote"])
        order.order_fills.update(
            {key: TradeUpdate.from_json(value) for key, value in data.get("order_fills", {}).items()}
        )
        order._nonce = data["nonce"]
        order._cancel_tx_hash = data["cancel_tx_hash"]
        order._gas_price = Decimal(data["gas_price"]) if data["gas_price"] != "None" else None
        order._creation_transaction_hash = data["creation_transaction_hash"]
        order.last_update_timestamp = data.get("last_update_timestamp", order.creation_timestamp)

        order.check_filled_condition()

        return order

    def to_json(self) -> Dict[str, Any]:
        """
        Returns this InFlightOrder as a JSON object.
        :return: JSON object
        """
        return {
            "client_order_id": self.client_order_id,
            "exchange_order_id": self.exchange_order_id,
            "trading_pair": self.trading_pair,
            "order_type": self.order_type.name,
            "trade_type": self.trade_type.name,
            "price": str(self.price),
            "amount": str(self.amount),
            "executed_amount_base": str(self.executed_amount_base),
            "executed_amount_quote": str(self.executed_amount_quote),
            "last_state": str(self.current_state.value),
            "leverage": str(self.leverage),
            "position": self.position.value,
            "creation_timestamp": self.creation_timestamp,
            "last_update_timestamp": self.last_update_timestamp,
            "order_fills": {key: fill.to_json() for key, fill in self.order_fills.items()},
            "nonce": self._nonce,
            "cancel_tx_hash": self._cancel_tx_hash,
            "creation_transaction_hash": self._creation_transaction_hash,
            "gas_price": str(self._gas_price),
        }


class GatewayPerpetualInFlightOrder(GatewayInFlightOrder):
    def build_order_created_message(self) -> str:
        return (
            f"Created {self.order_type.name.upper()} {self.trade_type.name.upper()} order "
            f"{self.client_order_id} for {self.amount} to {self.position.name.upper()} a {self.trading_pair} position."
        )



import binascii
import importlib
import logging
import shutil
from os import DirEntry, scandir
from os.path import exists, join
from typing import Any, Dict, List, Optional, Union, cast

import yaml

from hummingbot import root_path
from hummingbot.client.config.client_config_map import (
    AnonymizedMetricsDisabledMode,
    AnonymizedMetricsEnabledMode,
    ClientConfigMap,
    ColorConfigMap,
    DBOtherMode,
    DBSqliteMode,
    KillSwitchDisabledMode,
    KillSwitchEnabledMode,
)
from hummingbot.client.config.config_crypt import BaseSecretsManager, store_password_verification
from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.client.config.config_helpers import ClientConfigAdapter, save_to_yml
from hummingbot.client.config.security import Security
from hummingbot.client.settings import CLIENT_CONFIG_PATH, CONF_DIR_PATH, STRATEGIES_CONF_DIR_PATH
from hummingbot.strategy.avellaneda_market_making.avellaneda_market_making_config_map_pydantic import (
    AvellanedaMarketMakingConfigMap,
)
from hummingbot.strategy.cross_exchange_market_making.cross_exchange_market_making_config_map_pydantic import (
    CrossExchangeMarketMakingConfigMap,
)

encrypted_conf_prefix = "encrypted_"
encrypted_conf_postfix = ".json"
conf_dir_path = CONF_DIR_PATH
strategies_conf_dir_path = STRATEGIES_CONF_DIR_PATH


def migrate_configs(secrets_manager: BaseSecretsManager) -> List[str]:
    logging.getLogger().info("Starting conf migration.")
    errors = backup_existing_dir()
    if len(errors) == 0:
        errors = migrate_global_config()
        if len(errors) == 0:
            errors.extend(migrate_strategy_confs_paths())
            errors.extend(migrate_connector_confs(secrets_manager))
            store_password_verification(secrets_manager)
            logging.getLogger().info("\nConf migration done.")
    else:
        logging.getLogger().error("\nConf migration failed.")
    return errors


def migrate_non_secure_configs_only() -> List[str]:
    logging.getLogger().info("Starting strategies conf migration.")
    errors = backup_existing_dir()
    if len(errors) == 0:
        errors = migrate_global_config()
        if len(errors) == 0:
            errors.extend(migrate_strategy_confs_paths())
            logging.getLogger().info("\nConf migration done.")
    else:
        logging.getLogger().error("\nConf migration failed.")
    return errors


def backup_existing_dir() -> List[str]:
    errors = []
    if conf_dir_path.exists():
        backup_path = conf_dir_path.parent / "conf_backup"
        if backup_path.exists():
            errors = [
                (
                    f"\nBackup path {backup_path} already exists."
                    f"\nThe migration script cannot backup you existing"
                    f"\nconf files without overwriting that directory."
                    f"\nPlease remove it and run the script again."
                )
            ]
        else:
            shutil.copytree(conf_dir_path, backup_path)
            logging.getLogger().info(f"\nCreated a backup of your existing conf directory to {backup_path}")
    return errors


def migrate_global_config() -> List[str]:
    logging.getLogger().info("\nMigrating the global config...")
    global_config_path = CONF_DIR_PATH / "conf_global.yml"
    errors = []
    if global_config_path.exists():
        with open(str(global_config_path), "r") as f:
            data = yaml.safe_load(f)
        del data["template_version"]
        client_config_map = ClientConfigAdapter(ClientConfigMap())
        _migrate_global_config_modes(client_config_map, data)
        data.pop("kraken_api_tier", None)
        data.pop("key_file_path", None)
        keys = list(data.keys())
        for key in keys:
            if key in client_config_map.keys():
                _migrate_global_config_field(client_config_map, data, key)
        for key in data:
            logging.getLogger().warning(f"Global ConfigVar {key} was not migrated.")
        if len(errors) == 0:
            save_to_yml(CLIENT_CONFIG_PATH, client_config_map)
            global_config_path.unlink()
            logging.getLogger().info("\nSuccessfully migrated the global config.")
        else:
            errors = [f"client_config_map - {e}" for e in errors]
            logging.getLogger().error(f"The migration of the global config map failed with errors: {errors}")
    return errors


def _migrate_global_config_modes(client_config_map: ClientConfigAdapter, data: Dict):
    client_config_map: Union[ClientConfigAdapter, ClientConfigMap] = client_config_map  # for IDE autocomplete

    kill_switch_enabled = data.pop("kill_switch_enabled")
    kill_switch_rate = data.pop("kill_switch_rate")
    if kill_switch_enabled:
        client_config_map.kill_switch_mode = KillSwitchEnabledMode(kill_switch_rate=kill_switch_rate)
    else:
        client_config_map.kill_switch_mode = KillSwitchDisabledMode()

    _migrate_global_config_field(
        client_config_map.paper_trade, data, "paper_trade_exchanges"
    )
    _migrate_global_config_field(
        client_config_map.paper_trade, data, "paper_trade_account_balance"
    )
    db_engine = data.pop("db_engine")
    db_host = data.pop("db_host")
    db_port = data.pop("db_port")
    db_username = data.pop("db_username")
    db_password = data.pop("db_password")
    db_name = data.pop("db_name")
    if db_engine == "sqlite":
        client_config_map.db_mode = DBSqliteMode()
    else:
        client_config_map.db_mode = DBOtherMode(
            db_engine=db_engine,
            db_host=db_host,
            db_port=db_port,
            db_username=db_username,
            db_password=db_password,
            db_name=db_name,
        )

    _migrate_global_config_field(
        client_config_map.gateway, data, "gateway_api_port"
    )

    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_host"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_port"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_username"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_password"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_ssl"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_logger"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_notifier"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_commands"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_events"
    )
    _migrate_global_config_field(
        client_config_map.mqtt_bridge, data, "mqtt_autostart"
    )

    anonymized_metrics_enabled = data.pop("anonymized_metrics_enabled")
    anonymized_metrics_interval_min = data.pop("anonymized_metrics_interval_min")
    if anonymized_metrics_enabled:
        client_config_map.anonymized_metrics_mode = AnonymizedMetricsEnabledMode(
            anonymized_metrics_interval_min=anonymized_metrics_interval_min
        )
    else:
        client_config_map.anonymized_metrics_mode = AnonymizedMetricsDisabledMode()

    _migrate_global_config_field(
        client_config_map.global_token, data, "global_token", "global_token_name"
    )
    _migrate_global_config_field(
        client_config_map.global_token, data, "global_token_symbol"
    )

    _migrate_global_config_field(
        client_config_map.commands_timeout, data, "create_command_timeout"
    )
    _migrate_global_config_field(
        client_config_map.commands_timeout, data, "other_commands_timeout"
    )

    color_map: Union[ClientConfigAdapter, ColorConfigMap] = client_config_map.color
    _migrate_global_config_field(color_map, data, "top-pane", "top_pane")
    _migrate_global_config_field(color_map, data, "bottom-pane", "bottom_pane")
    _migrate_global_config_field(color_map, data, "output-pane", "output_pane")
    _migrate_global_config_field(color_map, data, "input-pane", "input_pane")
    _migrate_global_config_field(color_map, data, "logs-pane", "logs_pane")
    _migrate_global_config_field(color_map, data, "terminal-primary", "terminal_primary")
    _migrate_global_config_field(color_map, data, "primary-label", "primary_label")
    _migrate_global_config_field(color_map, data, "secondary-label", "secondary_label")
    _migrate_global_config_field(color_map, data, "success-label", "success_label")
    _migrate_global_config_field(color_map, data, "warning-label", "warning_label")
    _migrate_global_config_field(color_map, data, "info-label", "info_label")
    _migrate_global_config_field(color_map, data, "error-label", "error_label")

    balance_asset_limit = data.pop("balance_asset_limit")
    if balance_asset_limit is not None:
        exchanges = list(balance_asset_limit.keys())
        for e in exchanges:
            if balance_asset_limit[e] is None:
                balance_asset_limit.pop(e)
            else:
                assets = balance_asset_limit[e].keys()
                for a in assets:
                    if balance_asset_limit[e][a] is None:
                        balance_asset_limit[e].pop(a)
        client_config_map.balance_asset_limit = balance_asset_limit


def _migrate_global_config_field(
    cm: ClientConfigAdapter, global_config_data: Dict[str, Any], attr: str, cm_attr: Optional[str] = None
):
    value = global_config_data.pop(attr)
    cm_attr = cm_attr if cm_attr is not None else attr
    if value is not None:
        cm.setattr_no_validation(cm_attr, value)


def migrate_strategy_confs_paths():
    errors = []
    logging.getLogger().info("\nMigrating strategies...")
    for child in conf_dir_path.iterdir():
        if child.is_file() and child.name.endswith(".yml"):
            with open(str(child), "r") as f:
                conf = yaml.safe_load(f)
            if "strategy" in conf and _has_connector_field(conf):
                new_path = strategies_conf_dir_path / child.name
                child.rename(new_path)
                if conf["strategy"] == "avellaneda_market_making":
                    errors.extend(migrate_amm_confs(conf, new_path))
                elif conf["strategy"] == "cross_exchange_market_making":
                    errors.extend(migrate_xemm_confs(conf, new_path))
                logging.getLogger().info(f"Migrated conf for {conf['strategy']}")
    return errors


def migrate_amm_confs(conf, new_path) -> List[str]:
    execution_timeframe = conf.pop("execution_timeframe")
    if execution_timeframe == "infinite":
        conf["execution_timeframe_mode"] = {}
        conf.pop("start_time")
        conf.pop("end_time")
    elif execution_timeframe == "from_date_to_date":
        conf["execution_timeframe_mode"] = {
            "start_datetime": conf.pop("start_time"),
            "end_datetime": conf.pop("end_time"),
        }
    else:
        assert execution_timeframe == "daily_between_times"
        conf["execution_timeframe_mode"] = {
            "start_time": conf.pop("start_time"),
            "end_time": conf.pop("end_time"),
        }
    order_levels = int(conf.pop("order_levels"))
    if order_levels == 1:
        conf["order_levels_mode"] = {}
        conf.pop("level_distances")
    else:
        conf["order_levels_mode"] = {
            "order_levels": order_levels,
            "level_distances": conf.pop("level_distances")
        }
    hanging_orders_enabled = conf.pop("hanging_orders_enabled")
    if not hanging_orders_enabled:
        conf["hanging_orders_mode"] = {}
        conf.pop("hanging_orders_cancel_pct")
    else:
        conf["hanging_orders_mode"] = {
            "hanging_orders_cancel_pct": conf.pop("hanging_orders_cancel_pct")
        }
    if "template_version" in conf:
        conf.pop("template_version")
    try:
        config_map = ClientConfigAdapter(AvellanedaMarketMakingConfigMap(**conf))
        save_to_yml(new_path, config_map)
        errors = []
    except Exception as e:
        logging.getLogger().error(str(e))
        errors = [str(e)]
    return errors


def migrate_xemm_confs(conf, new_path) -> List[str]:
    if "active_order_canceling" in conf:
        if conf["active_order_canceling"]:
            conf["order_refresh_mode"] = {}
        else:
            conf["order_refresh_mode"] = {
                "cancel_order_threshold": conf["cancel_order_threshold"],
                "limit_order_min_expiration": conf["limit_order_min_expiration"]
            }
        conf.pop("active_order_canceling")
        conf.pop("cancel_order_threshold")
        conf.pop("limit_order_min_expiration")
    if "use_oracle_conversion_rate" in conf:
        if conf["use_oracle_conversion_rate"]:
            conf["conversion_rate_mode"] = {}
        else:
            conf["conversion_rate_mode"] = {
                "taker_to_maker_base_conversion_rate": conf["taker_to_maker_base_conversion_rate"],
                "taker_to_maker_quote_conversion_rate": conf["taker_to_maker_quote_conversion_rate"]
            }
        conf.pop("use_oracle_conversion_rate")
        conf.pop("taker_to_maker_base_conversion_rate")
        conf.pop("taker_to_maker_quote_conversion_rate")
    if "template_version" in conf:
        conf.pop("template_version")
    try:
        config_map = ClientConfigAdapter(CrossExchangeMarketMakingConfigMap(**conf))
        save_to_yml(new_path, config_map)
        errors = []
    except Exception as e:
        logging.getLogger().error(str(e))
        errors = [str(e)]
    return errors


def _has_connector_field(conf: Dict) -> bool:
    return (
        "exchange" in conf
        or "connector_1" in conf  # amm arb
        or "primary_market" in conf  # arbitrage
        or "secondary_exchange" in conf  # celo arb
        or "maker_market" in conf  # XEMM
        or "market" in conf  # dev simple trade
        or "maker_exchange" in conf  # hedge
        or "spot_connector" in conf  # spot-perp arb
        or "connector" in conf  # twap
    )


def migrate_connector_confs(secrets_manager: BaseSecretsManager):
    logging.getLogger().info("\nMigrating connector secure keys...")
    errors = []
    Security.secrets_manager = secrets_manager
    connector_exceptions = ["paper_trade"]
    type_dirs: List[DirEntry] = [
        cast(DirEntry, f) for f in
        scandir(f"{root_path() / 'hummingbot' / 'connector'}")
        if f.is_dir()
    ]
    for type_dir in type_dirs:
        connector_dirs: List[DirEntry] = [
            cast(DirEntry, f) for f in scandir(type_dir.path)
            if f.is_dir() and exists(join(f.path, "__init__.py"))
        ]
        for connector_dir in connector_dirs:
            if connector_dir.name.startswith("_") or connector_dir.name in connector_exceptions:
                continue
            try:
                suffix = "data_types" if connector_dir.name == "celo" else "utils"
                util_module_path: str = (
                    f"hummingbot.connector.{type_dir.name}.{connector_dir.name}.{connector_dir.name}_{suffix}"
                )
                util_module = importlib.import_module(util_module_path)
                config_keys = getattr(util_module, "KEYS", None)
                if config_keys is not None:
                    errors.extend(_maybe_migrate_encrypted_confs(config_keys))
                other_domains = getattr(util_module, "OTHER_DOMAINS", [])
                for domain in other_domains:
                    config_keys = getattr(util_module, "OTHER_DOMAINS_KEYS")[domain]
                    if config_keys is not None:
                        errors.extend(_maybe_migrate_encrypted_confs(config_keys))
            except ModuleNotFoundError:
                continue
    return errors


def _maybe_migrate_encrypted_confs(config_keys: BaseConnectorConfigMap) -> List[str]:
    cm = ClientConfigAdapter(config_keys)
    found_one = False
    files_to_remove = []
    missing_fields = []
    for el in cm.traverse():
        if el.client_field_data is not None:
            key_path = conf_dir_path / f"{encrypted_conf_prefix}{el.attr}{encrypted_conf_postfix}"
            if key_path.exists():
                with open(key_path, 'r') as f:
                    json_str = f.read()
                value = binascii.hexlify(json_str.encode()).decode()
                if not el.client_field_data.is_secure:
                    value = Security.secrets_manager.decrypt_secret_value(el.attr, value)
                cm.setattr_no_validation(el.attr, value)
                files_to_remove.append(key_path)
                found_one = True
            else:
                missing_fields.append(el.attr)
    errors = []
    if found_one:
        if len(missing_fields) != 0:
            errors = [f"{config_keys.connector} - missing fields: {missing_fields}"]
        else:
            Security.update_secure_config(cm)
            logging.getLogger().info(f"Migrated secure keys for {config_keys.connector}")
        for f in files_to_remove:
            f.unlink()
    return errors



import time
from typing import Any, Dict, Optional

from hummingbot.connector.exchange.ascend_ex import ascend_ex_constants as CONSTANTS
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest
from hummingbot.core.web_assistant.rest_pre_processors import RESTPreProcessorBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class AscendExRESTPreProcessor(RESTPreProcessorBase):
    async def pre_process(self, request: RESTRequest) -> RESTRequest:
        if request.headers is None:
            request.headers = {}
        # Generates generic headers required by AscendEx
        headers_generic = {}
        headers_generic["Accept"] = "application/json"
        headers_generic["Content-Type"] = "application/json"
        # Headers signature to identify user as an HB liquidity provider.
        request.headers = dict(
            list(request.headers.items()) + list(headers_generic.items()) + list(get_hb_id_headers().items())
        )
        return request


def get_hb_id_headers() -> Dict[str, Any]:
    """
    Headers signature to identify user as an HB liquidity provider.

    :return: a custom HB signature header
    """
    return {
        "request-source": "hummingbot-liq-mining",
    }


def public_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided private REST endpoint
    :param path_url: a private REST endpoint
    :param domain: domain to connect to
    :return: the full URL to the endpoint
    """
    return CONSTANTS.PUBLIC_REST_URL + path_url


def private_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided private REST endpoint
    :param path_url: a private REST endpoint
    :param domain: the domain to connect to
    :return: None, we use overwrite_url instead
    """
    return CONSTANTS.PRIVATE_REST_URL + path_url


def build_api_factory(
    throttler: Optional[AsyncThrottler] = None,
    domain: str = CONSTANTS.DEFAULT_DOMAIN,
    auth: Optional[AuthBase] = None,
) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    api_factory = WebAssistantsFactory(throttler=throttler, auth=auth, rest_pre_processors=[AscendExRESTPreProcessor()])
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def get_current_server_time(
    throttler: Optional[AsyncThrottler] = None,
    domain: str = CONSTANTS.DEFAULT_DOMAIN,
) -> int:
    return int(time.time() * 1e3)
